import h5py
import numpy as np
from pymatgen import Structure

from amset.constants import str_to_spin

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


def write_coefficients(coeffs, gpoints, kpoints, structure, filename="coeffs.h5"):
    with h5py.File(filename, "w") as f:
        for spin, spin_coeffs in coeffs.items():
            name = "coefficients_{}".format(spin.name)
            dset = f.create_dataset(
                name, spin_coeffs.shape, compression="gzip", dtype=np.complex
            )
            dset[...] = spin_coeffs

        f["structure"] = np.string_(structure.to_json())
        f["kpoints"] = kpoints
        f["gpoints"] = gpoints


def load_coefficients(filename):
    coeffs = {}
    with h5py.File(filename, "r") as f:
        coeff_keys = [k for k in list(f.keys()) if "coefficients" in k]
        for key in coeff_keys:
            spin = str_to_spin[key.split("_")[1]]
            coeffs[spin] = np.array(f[key])

        structure_str = np.string_(np.array(f["structure"])).decode()
        structure = Structure.from_str(structure_str, fmt="json")
        kpoints = np.array(f["kpoints"])
        if "gpoints" in f:
            gpoints = np.array(f["gpoints"])
        else:
            gpoints = None

    return coeffs, gpoints, kpoints, structure



from pathlib import Path

import click
from click import argument, option

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

from amset.tools.common import image_type, path_type, zero_weighted_type

kpaths = click.Choice(["pymatgen", "seekpath"], case_sensitive=False)

_symprec = 0.01  # redefine symprec to avoid loading constants from file
_dos_estep = 0.01  # redefine symprec to avoid loading constants from file
_interpolation_factor = 5


def _all_or_int(value):
    if value == "all":
        return None
    elif value is None:
        return 0
    else:
        return int(value)


def _parse_kpoints(kpoints):
    return [
        [list(map(float, kpt.split())) for kpt in kpts.split(",")]
        for kpts in kpoints.split("|")
    ]


def _parse_kpoint_labels(labels):
    return [path.split(",") for path in labels.split("|")]


@click.group()
def plot():
    """
    Plot AMSET results, including scattering rates and band structures
    """
    pass


@plot.command()
@argument("filename", type=path_type)
@option("-t", "--temperature", default=0, help="temperature index [default: 0]")
@option("-d", "--doping", default=0, help="doping index [default: 0]")
@option("-l", "--line-density", default=100.0, help="band structure line density")
@option("-p", "--prefix", help="output filename prefix")
@option("--emin", help="minimum energy limit", type=float)
@option("--emax", help="maximum energy limit", type=float)
@option(
    "--amin",
    default=5e-5,
    help="minimum spectral weight for normalizing linewidth in 1/meV",
)
@option(
    "--amax",
    default=1e-1,
    help="maximum spectral weight for normalizing linewidth in 1/meV",
)
@option("--cmap", default="viridis", help="matplotlib colormap to use")
@option("--no-colorbar", default=True, is_flag=True, help="don't add a colorbar")
@option("--symprec", type=float, default=_symprec, help="interpolation factor")
@option("--kpath", type=kpaths, help="k-point path type")
@option(
    "--kpoints",
    type=_parse_kpoints,
    metavar="K",
    help="manual k-points list [e.g. '0 0 0, 0.5 0 0']",
)
@option(
    "--labels",
    type=_parse_kpoint_labels,
    metavar="L",
    help=r"labels for manual kpoints [e.g. '\Gamma,X']",
)
@option("--interpolation-factor", default=1, help="BoltzTraP interpolation factor")
@option("--width", default=3.2, help="figure width [default: 6]")
@option("--height", default=3.2, help="figure height [default: 6]")
@option("--directory", type=path_type, help="file output directory")
@option("--image-format", default="pdf", type=image_type, help="image format")
@option("--style", help="matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
def lineshape(filename, **kwargs):
    """
    Plot band structures with electron lineshape
    """
    from amset.plot.lineshape import LineshapePlotter

    plotter = LineshapePlotter(
        filename, print_log=True, interpolation_factor=kwargs["interpolation_factor"]
    )

    kpath = get_kpath(
        plotter.structure,
        mode=kwargs["kpath"],
        symprec=kwargs["symprec"],
        kpt_list=kwargs["kpoints"],
        labels=kwargs["labels"],
    )

    plt = plotter.get_plot(
        kwargs["doping"],
        kwargs["temperature"],
        line_density=kwargs["line_density"],
        emin=kwargs["emin"],
        emax=kwargs["emax"],
        amin=kwargs["amin"],
        amax=kwargs["amax"],
        cmap=kwargs["cmap"],
        colorbar=kwargs["no_colorbar"],
        width=kwargs["width"],
        height=kwargs["height"],
        style=kwargs["style"],
        no_base_style=kwargs["no_base_style"],
        kpath=kpath,
    )

    save_plot(
        plt, "lineshape", kwargs["directory"], kwargs["prefix"], kwargs["image_format"]
    )
    return plt


@plot.command()
@argument("filename", type=path_type)
@option("-l", "--line-density", default=100.0, help="band structure line density")
@option("--emin", default=-6.0, help="minimum energy limit")
@option("--emax", default=6.0, help="maximum energy limit")
@option("--symprec", type=float, default=_symprec, help="interpolation factor")
@option("--kpath", type=kpaths, help="k-point path type")
@option(
    "--kpoints",
    type=_parse_kpoints,
    metavar="K",
    help="manual k-points list [e.g. '0 0 0, 0.5 0 0']",
)
@option(
    "--labels",
    type=_parse_kpoint_labels,
    metavar="L",
    help=r"labels for manual kpoints [e.g. '\Gamma,X']",
)
@option(
    "--interpolation-factor",
    default=_interpolation_factor,
    type=float,
    help="BoltzTraP interpolation factor",
)
@option("--energy-cutoff", type=float, help="interpolation energy cutoff in eV")
@option(
    "-z",
    "--zero-weighted-kpoints",
    help="how to process zero-weighted k-points",
    type=zero_weighted_type,
)
@option("--plot-dos", is_flag=True, help="whether to also plot the density of states")
@option(
    "--dos-kpoints",
    default="50",
    help="k-point length cutoff or mesh for density of states",
)
@option("--dos-estep", default=_dos_estep, help="dos energy step size")
@option("--dos-aspect", default=3.0, help="aspect ratio for the density of states")
@option(
    "--no-zero-to-efermi",
    "zero_to_efermi",
    is_flag=True,
    default=True,
    help="don't set the Fermi level to zero",
)
@option("--vbm-cbm-marker", is_flag=True, help="add a marker at the CBM and VBM")
@option("--width", default=6.0, help="figure width [default: 6]")
@option("--height", default=6.0, help="figure height [default: 6]")
@option("-p", "--prefix", help="output filename prefix")
@option("--directory", type=path_type, help="file output directory")
@option("--image-format", default="pdf", type=image_type, help="image format")
@option("--style", help="path to matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
def band(filename, **kwargs):
    """
    Plot interpolate band structure from vasprun file
    """
    from amset.constants import defaults
    from amset.plot.electronic_structure import ElectronicStructurePlotter

    zwk_mode = kwargs.pop("zero_weighted_kpoints")
    if not zwk_mode:
        zwk_mode = defaults["zero_weighted_kpoints"]

    plotter_kwargs = {
        "print_log": True,
        "interpolation_factor": kwargs["interpolation_factor"],
        "symprec": kwargs["symprec"],
        "energy_cutoff": kwargs["energy_cutoff"],
    }
    if not is_vasprun_file(filename):
        click.Abort("Unrecognised filetype, expecting a vasprun.xml file.")

    plotter = ElectronicStructurePlotter.from_vasprun(
        filename, zero_weighted_kpoints=zwk_mode, **plotter_kwargs
    )
    kpath = get_kpath(
        plotter.structure,
        mode=kwargs["kpath"],
        symprec=kwargs["symprec"],
        kpt_list=kwargs["kpoints"],
        labels=kwargs["labels"],
    )

    dos_kpoints = _get_dos_kpoints(plotter.structure, kwargs["dos_kpoints"])

    plt = plotter.get_plot(
        plot_band_structure=True,
        plot_dos=kwargs["plot_dos"],
        line_density=kwargs["line_density"],
        dos_kpoints=dos_kpoints,
        dos_estep=kwargs["dos_estep"],
        dos_aspect=kwargs["dos_aspect"],
        zero_to_efermi=kwargs["zero_to_efermi"],
        vbm_cbm_marker=kwargs["vbm_cbm_marker"],
        emin=kwargs["emin"],
        emax=kwargs["emax"],
        width=kwargs["width"],
        height=kwargs["height"],
        style=kwargs["style"],
        no_base_style=kwargs["no_base_style"],
        kpath=kpath,
    )

    save_plot(
        plt, "band", kwargs["directory"], kwargs["prefix"], kwargs["image_format"]
    )
    return plt


@plot.command()
@argument("filename", type=path_type)
@option("--ymin", default=None, type=float, help="minimum y-axis limit")
@option("--ymax", default=None, type=float, help="maximum y-axis limit")
@option("-d", "--doping-idx", metavar="N", help="doping index to plot")
@option("-t", "--temperature-idx", metavar="N", help="temperature index to plot")
@option(
    "-s",
    "--separate-rates",
    is_flag=True,
    default=False,
    help="whether to separate scattering mechanisms",
)
@option("-p", "--prefix", help="output filename prefix")
@option("--directory", type=path_type, help="file output directory")
@option("--image-format", default="pdf", type=image_type, help="image format")
@option("--style", help="path to matplotlib style specification")
@option("--no-base-style", default=False, is_flag=True, help="don't apply base style")
def rates(filename, **kwargs):
    """
    Plot scattering rates
    """
    from amset.plot.rates import RatesPlotter

    save_kwargs = [kwargs.pop(d) for d in ["directory", "prefix", "image_format"]]

    kwargs["doping_idx"] = _all_or_int(kwargs["doping_idx"])
    kwargs["temperature_idx"] = _all_or_int(kwargs["temperature_idx"])

    plotter = RatesPlotter(filename)
    plt = plotter.get_plot(**kwargs)
    plt.subplots_adjust(hspace=0.3, wspace=0.3)
    save_plot(plt, "rates", *save_kwargs)
    return plt


def save_plot(plt, name, directory, prefix, image_format):
    if prefix:
        prefix += "_"
    else:
        prefix = ""
    filename = Path("{}{}.{}".format(prefix, name, image_format))

    if directory:
        filename = directory / filename

    plt.savefig(filename, format=image_format, bbox_inches="tight", dpi=400)


def is_vasprun_file(filename):
    return "xml" in filename


def get_kpath(structure, mode="pymatgen", symprec=_symprec, kpt_list=None, labels=None):
    r"""Get a Kpath object

    If a manual list of kpoints is supplied using the `kpt_list`
    variable, the `mode` option will be ignored.

    Args:
        structure: The structure.
        mode: Method used for calculating the
            high-symmetry path. The options are:

            pymatgen
                Use the paths from pymatgen.

            seekpath
                Use the paths from SeeK-path.

        symprec: The tolerance for determining the crystal symmetry.
        kpt_list: List of k-points to use, formatted as a list of subpaths, each
            containing a list of fractional k-points. For example:

            ```
                [ [[0., 0., 0.], [0., 0., 0.5]], [[0.5, 0., 0.], [0.5, 0.5, 0.]] ]
            ```

            will return points along `0 0 0 -> 0 0 1/2 | 1/2 0 0 -> 1/2 1/2 0`
        labels: The k-point labels. These should be provided as a list of strings for
            each subpath of the overall path. For example::

            ```
                [ ['Gamma', 'Z'], ['X', 'M'] ]
            ```

            combined with the above example for `kpt_list` would indicate the path:
            `Gamma -> Z | X -> M`. If no labels are provided, letters from A -> Z will
            be used instead.

    Returns:
        A Kpath object.
    """
    from sumo.symmetry import CustomKpath, PymatgenKpath, SeekpathKpath

    if kpt_list:
        kpath = CustomKpath(structure, kpt_list, labels, symprec=symprec)
    elif mode == "seekpath":
        kpath = SeekpathKpath(structure, symprec=symprec)
    else:
        kpath = PymatgenKpath(structure, symprec=symprec)

    return kpath


def _get_dos_kpoints(structure, kpoints):
    from amset.electronic_structure.kpoints import get_kpoint_mesh

    if kpoints is None:
        return None

    splits = kpoints.split()
    if len(splits) == 1:
        return get_kpoint_mesh(structure, float(splits[0]))
    else:
        return list(map(int, splits))



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Tuple

import numpy as np
from pymatgen import Spin
from tabulate import tabulate

from amset.constants import (
    boltzmann_au,
    coulomb_to_au,
    ev_to_hartree,
    gpa_to_au,
    m_to_bohr,
    s_to_au,
)
from amset.core.data import AmsetData, check_nbands_equal
from amset.interpolation.deformation import DeformationPotentialInterpolator
from amset.scattering.common import calculate_inverse_screening_length_sq

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class AbstractElasticScattering(ABC):

    name: str
    required_properties: Tuple[str]

    def __init__(self, properties, doping, temperatures, nbands):
        self.properties = properties
        self.doping = doping
        self.temperatures = temperatures
        self.nbands = nbands
        self.spins = list(nbands.keys())

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
        )

    @abstractmethod
    def prefactor(self, spin: Spin, b_idx: int):
        pass

    @abstractmethod
    def factor(
        self,
        unit_q: np.array,
        norm_q_sq: np.ndarray,
        spin: Spin,
        band_idx: int,
        kpoint: np.ndarray,
        velocity: np.ndarray,
    ):
        pass

    def to_reference(self):
        return self.properties, self.doping, self.temperatures, self.nbands

    @classmethod
    def from_reference(cls, properties, doping, temperatures, nbands):
        return cls(properties, doping, temperatures, nbands)

    @classmethod
    def get_properties(cls, materials_properties):
        return {p: materials_properties[p] for p in cls.required_properties}

    @staticmethod
    def get_nbands(amset_data):
        return {s: len(amset_data.energies[s]) for s in amset_data.spins}


class AcousticDeformationPotentialScattering(AbstractElasticScattering):

    name = "ADP"
    required_properties = ("deformation_potential", "elastic_constant")

    def __init__(
        self,
        properties,
        doping,
        temperatures,
        nbands,
        deformation_potential,
        vb_idx,
        is_metal,
        fermi_levels,
    ):
        super().__init__(properties, doping, temperatures, nbands)
        self._prefactor = boltzmann_au * s_to_au
        self.elastic_constant = self.properties["elastic_constant"] * gpa_to_au
        self.deformation_potential = deformation_potential
        self.vb_idx = vb_idx
        self.is_metal = is_metal
        self.fermi_levels = fermi_levels

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        vb_idx = amset_data.vb_idx
        is_metal = amset_data.is_metal

        deformation_potential = materials_properties["deformation_potential"]
        if isinstance(deformation_potential, (str, Path)):
            deformation_potential = DeformationPotentialInterpolator.from_file(
                deformation_potential, scale=ev_to_hartree
            )
            equal = check_nbands_equal(deformation_potential, amset_data)
            if not equal:
                raise RuntimeError(
                    "Deformation potential file does not contain the correct number of"
                    " bands\nEnsure it was generated using the same energy_cutoff as "
                    "this AMSET run."
                )

        elif is_metal and isinstance(deformation_potential, tuple):
            logger.warning(
                "System is metallic but deformation potentials for both "
                "the valence and conduction bands have been set... using the "
                "valence band potential for all bands"
            )
            deformation_potential = deformation_potential[0] * ev_to_hartree

        elif is_metal:
            deformation_potential = deformation_potential * ev_to_hartree

        elif not is_metal and not isinstance(deformation_potential, tuple):
            logger.warning(
                "System is semiconducting but only one deformation "
                "potential has been set... using this potential for all bands."
            )
            deformation_potential = (
                deformation_potential * ev_to_hartree,
                deformation_potential * ev_to_hartree,
            )

        else:
            deformation_potential = (
                deformation_potential[0] * ev_to_hartree,
                deformation_potential[1] * ev_to_hartree,
            )
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            deformation_potential,
            vb_idx,
            is_metal,
            amset_data.fermi_levels,
        )

    def prefactor(self, spin: Spin, b_idx: int):
        prefactor = (
            self._prefactor
            * self.temperatures[None, :]
            * np.ones((len(self.doping), len(self.temperatures)))
        )
        return prefactor

    def factor(
        self,
        unit_q: np.array,
        norm_q_sq: np.ndarray,
        spin: Spin,
        band_idx: int,
        kpoint: np.ndarray,
        velocity: np.ndarray,
    ):
        christoffel_tensors = get_christoffel_tensors(self.elastic_constant, unit_q)
        (
            (c_trans_a, c_trans_b, c_long),
            (v_trans_a, v_trans_b, v_long),
        ) = solve_christoffel_equation(christoffel_tensors)
        if isinstance(self.deformation_potential, DeformationPotentialInterpolator):
            deform = self.deformation_potential.interpolate(spin, [band_idx], [kpoint])
            deform = np.abs(deform[0])
            deform += np.outer(velocity, velocity)  # velocity correction
            strain_long, strain_trans_a, strain_trans_b = prepare_acoustic_strains(
                unit_q, v_long, v_trans_a, v_trans_b
            )
            factor = (
                np.tensordot(strain_long, deform) ** 2 / c_long
                + np.tensordot(strain_trans_a, deform) ** 2 / c_trans_a
                + np.tensordot(strain_trans_b, deform) ** 2 / c_trans_b
            )
        elif self.is_metal:
            factor = self.deformation_potential ** 2 / c_long
        else:
            def_idx = 1 if band_idx > self.vb_idx[spin] else 0
            factor = self.deformation_potential[def_idx] ** 2 / c_long

        return factor[None, None] * np.ones(self.fermi_levels.shape + norm_q_sq.shape)

    def to_reference(self):
        base_reference = super().to_reference()
        if isinstance(self.deformation_potential, DeformationPotentialInterpolator):
            deformation_reference = self.deformation_potential.to_reference()
            is_interpolator = True
        else:
            deformation_reference = self.deformation_potential
            is_interpolator = False
        return base_reference + (
            deformation_reference,
            self.vb_idx,
            self.is_metal,
            self.fermi_levels,
            is_interpolator,
        )

    @classmethod
    def from_reference(
        cls,
        properties,
        doping,
        temperatures,
        nbands,
        deformation_reference,
        vb_idx,
        is_metal,
        fermi_levels,
        is_interpolator,
    ):
        if is_interpolator:
            deformation_potential = DeformationPotentialInterpolator.from_reference(
                *deformation_reference
            )
        else:
            deformation_potential = deformation_reference
        return cls(
            properties,
            doping,
            temperatures,
            nbands,
            deformation_potential,
            vb_idx,
            is_metal,
            fermi_levels,
        )


def prepare_acoustic_strains(unit_q, v_long, v_trans_a, v_trans_b):
    # orient v_long and unit_q to face the same direction
    # the einsum is just pairwise dot product along the first axis
    sign = np.sign(np.einsum("ij,ij->i", unit_q, v_long))[:, None]
    v_long *= sign
    v_trans_a *= sign
    v_trans_b *= sign

    strain_long = get_unit_strain_tensors(unit_q, v_long)
    strain_trans_a = get_unit_strain_tensors(unit_q, v_trans_a)
    strain_trans_b = get_unit_strain_tensors(unit_q, v_trans_b)
    return strain_long, strain_trans_a, strain_trans_b


def get_christoffel_tensors(elastic_constant, unit_q):
    return np.einsum("ijkl,ni,nl->njk", elastic_constant, unit_q, unit_q)


def solve_christoffel_equation(christoffel_tensors):
    eigenvalues, eigenvectors = np.linalg.eigh(christoffel_tensors)
    return eigenvalues.T, eigenvectors.transpose(2, 0, 1)


def get_unit_strain_tensors(propagation_vectors, polarization_vectors):
    return propagation_vectors[:, :, None] * polarization_vectors[:, None, :]


class IonizedImpurityScattering(AbstractElasticScattering):
    name = "IMP"
    required_properties = ("acceptor_charge", "donor_charge", "static_dielectric")

    def __init__(
        self,
        properties,
        doping,
        temperatures,
        nbands,
        impurity_concentration,
        inverse_screening_length_sq,
    ):
        super().__init__(properties, doping, temperatures, nbands)
        self._prefactor = impurity_concentration * s_to_au * np.pi
        self.inverse_screening_length_sq = inverse_screening_length_sq

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        from amset.constants import bohr_to_cm

        avg_diel = np.linalg.eigvalsh(materials_properties["static_dielectric"]).mean()
        inverse_screening_length_sq = calculate_inverse_screening_length_sq(
            amset_data, avg_diel
        )

        imp_info = []
        impurity_concentration = np.zeros(amset_data.fermi_levels.shape)
        for n, t in np.ndindex(inverse_screening_length_sq.shape):
            n_conc = np.abs(amset_data.electron_conc[n, t])
            p_conc = np.abs(amset_data.hole_conc[n, t])

            impurity_concentration[n, t] = (
                n_conc * materials_properties["donor_charge"] ** 2
                + p_conc * materials_properties["acceptor_charge"] ** 2
            )
            imp_info.append(
                (
                    amset_data.doping[n] * (1 / bohr_to_cm) ** 3,
                    amset_data.temperatures[t],
                    inverse_screening_length_sq[n, t],
                    impurity_concentration[n, t] * (1 / bohr_to_cm) ** 3,
                )
            )
        logger.info("Inverse screening length (β) and impurity concentration (Nᵢᵢ):")
        table = tabulate(
            imp_info,
            headers=("conc [cm⁻³]", "temp [K]", "β² [a₀⁻²]", "Nᵢᵢ [cm⁻³]"),
            numalign="right",
            stralign="center",
            floatfmt=(".2e", ".1f", ".2e", ".2e"),
        )
        logger.info(table)
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            impurity_concentration,
            inverse_screening_length_sq,
        )

    def prefactor(self, spin: Spin, b_idx: int):
        # need to return prefactor with shape (nspins, ndops, ntemps, nbands)
        return self._prefactor

    def factor(
        self,
        unit_q: np.array,
        norm_q_sq: np.ndarray,
        spin: Spin,
        band_idx: int,
        kpoint: np.ndarray,
        velocity: np.ndarray,
    ):
        static_tensor = self.properties["static_dielectric"] / (4 * np.pi)
        static_diel = np.einsum("ij,ij->i", unit_q, np.dot(static_tensor, unit_q.T).T)
        diel_factor = (1 / static_diel) ** 2
        return (
            diel_factor[None, None]
            / (norm_q_sq[None, None] + self.inverse_screening_length_sq[..., None]) ** 2
        )


class PiezoelectricScattering(AbstractElasticScattering):

    name = "PIE"
    required_properties = (
        "piezoelectric_constant",
        "elastic_constant",
        "high_frequency_dielectric",
        "free_carrier_screening",
    )

    def __init__(
        self,
        properties,
        doping,
        temperatures,
        nbands,
        piezoelectric_constant,
        inverse_screening_length_sq,
    ):
        super().__init__(properties, doping, temperatures, nbands)
        self.piezoelectric_constant = piezoelectric_constant
        self.inverse_screening_length_sq = inverse_screening_length_sq

        self._prefactor = self.temperatures[None, :] * boltzmann_au * s_to_au
        self._shape = np.ones((len(self.doping), len(self.temperatures)))
        self.elastic_constant = self.properties["elastic_constant"] * gpa_to_au

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        # convert dielectric to atomic units
        shape = (len(amset_data.doping), len(amset_data.temperatures))
        e = materials_properties["piezoelectric_constant"]
        e *= coulomb_to_au / m_to_bohr ** 2  # convert to atomic units
        dielectric = materials_properties["high_frequency_dielectric"] / (4 * np.pi)
        inv_dielectric = np.linalg.inv(dielectric)

        # use h piezoelectric coefficient (Stress-Voltage)
        piezoelectric_constant = np.einsum("mn,mkl->nkl", inv_dielectric, e)

        if materials_properties["free_carrier_screening"]:
            avg_diel = np.linalg.eigvalsh(
                materials_properties["high_frequency_dielectric"]
            ).mean()
            inverse_screening_length_sq = calculate_inverse_screening_length_sq(
                amset_data, avg_diel
            )
        else:
            # fill with small value for numerical convergence
            inverse_screening_length_sq = np.full(shape, 1e-12)
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            piezoelectric_constant,
            inverse_screening_length_sq,
        )

    def prefactor(self, spin: Spin, b_idx: int):
        # need to return prefactor with shape (ndops, ntemps)
        return self._prefactor * np.ones(self._shape)

    def factor(
        self,
        unit_q: np.array,
        norm_q_sq: np.ndarray,
        spin: Spin,
        band_idx: int,
        kpoint: np.ndarray,
        velocity: np.ndarray,
    ):
        christoffel_tensors = get_christoffel_tensors(self.elastic_constant, unit_q)
        (
            (c_trans_a, c_trans_b, c_long),
            (v_trans_a, v_trans_b, v_long),
        ) = solve_christoffel_equation(christoffel_tensors)

        strain_long, strain_trans_a, strain_trans_b = prepare_acoustic_strains(
            unit_q, v_long, v_trans_a, v_trans_b
        )
        qh = np.einsum("ijk,nj->nik", self.piezoelectric_constant, unit_q)

        # einsum is double dot product along first axis
        factor = (
            np.einsum("nij,nij->n", strain_long, qh) ** 2 / c_long
            + np.einsum("nij,nij->n", strain_trans_a, qh) ** 2 / c_trans_a
            + np.einsum("nij,nij->n", strain_trans_b, qh) ** 2 / c_trans_b
        )

        return (
            factor[None, None]
            * np.ones(self._shape + norm_q_sq.shape)
            / (norm_q_sq[None, None, :] + self.inverse_screening_length_sq[..., None])
        )



"""Band structure interpolation using BolzTraP2."""

import logging
import multiprocessing
import time
from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from BoltzTraP2 import fite, sphere
from monty.json import MSONable
from pymatgen.electronic_structure.bandstructure import (
    BandStructure,
    BandStructureSymmLine,
)
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.dos import Dos
from pymatgen.io.ase import AseAtomsAdaptor
from sumo.symmetry import Kpath, PymatgenKpath

from amset.constants import angstrom_to_bohr, au_to_s, bohr_to_cm
from amset.constants import defaults as defaults
from amset.constants import ev_to_hartree, hartree_to_ev, numeric_types, spin_name
from amset.core.data import AmsetData
from amset.electronic_structure.common import (
    get_atomic_structure,
    get_ibands,
    get_vb_idx,
)
from amset.electronic_structure.dos import FermiDos
from amset.electronic_structure.kpoints import (
    get_kpoints_tetrahedral,
    sort_boltztrap_to_spglib,
)
from amset.electronic_structure.symmetry import (
    get_symmetry_equivalent_kpoints,
    similarity_transformation,
)
from amset.electronic_structure.tetrahedron import TetrahedralBandStructure
from amset.interpolation.boltztrap import get_bands_fft
from amset.log import log_list, log_time_taken

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class Interpolator(MSONable):
    """Class to interpolate band structures based on BoltzTraP2.

    Details of the interpolation method are available in:

    3. Madsen, G. K. & Singh, D. J. Computer Physics Communications 175, 67–71
       (2006)
    3. Madsen, G. K., Carrete, J., and Verstraete, M. J. Computer Physics
        Communications 231, 140–145 (2018)

    Args:
        band_structure: A pymatgen band structure object.
        num_electrons: The number of electrons in the system.
        interpolation_factor: Factor used to determine the accuracy of the
            band structure interpolation. Also controls the k-point mesh density
            for :meth:`Interpolater.get_amset_data`.
        soc: Whether the system was calculated using spin–orbit coupling.
        magmom: The magnetic moments for each atom.
        mommat: The band structure derivatives.
        interpolate_projections: Whether to interpolate the band structure
            projections.
    """

    def __init__(
        self,
        band_structure: BandStructure,
        num_electrons: int,
        interpolation_factor: float = defaults["interpolation_factor"],
        soc: bool = False,
        magmom: Optional[np.ndarray] = None,
        mommat: Optional[np.ndarray] = None,
        other_properties: Dict[Spin, Dict[str, np.ndarray]] = None,
    ):
        self._band_structure = band_structure
        self._num_electrons = num_electrons
        self._soc = soc
        self._spins = self._band_structure.bands.keys()
        self._other_properties = other_properties
        self.interpolation_factor = interpolation_factor
        self._lattice_matrix = (
            band_structure.structure.lattice.matrix.T * angstrom_to_bohr
        )
        self._coefficients = {}
        self._other_coefficients = defaultdict(dict)

        kpoints = np.array([k.frac_coords for k in band_structure.kpoints])
        atoms = AseAtomsAdaptor.get_atoms(band_structure.structure)

        logger.info("Getting band interpolation coefficients")

        t0 = time.perf_counter()
        self._equivalences = sphere.get_equivalences(
            atoms, magmom, kpoints.shape[0] * interpolation_factor
        )

        # get the interpolation mesh used by BoltzTraP2
        self.interpolation_mesh = (
            2 * np.max(np.abs(np.vstack(self._equivalences)), axis=0) + 1
        )

        for spin in self._spins:
            energies = band_structure.bands[spin] * ev_to_hartree
            data = DFTData(kpoints, energies, self._lattice_matrix, mommat=mommat)
            self._coefficients[spin] = fite.fitde3D(data, self._equivalences)

        log_time_taken(t0)

        t0 = time.perf_counter()
        if self._other_properties:
            logger.info("Getting additional interpolation coefficients")

            for spin in self._spins:
                for label, prop in self._other_properties[spin].items():
                    data = DFTData(kpoints, prop, self._lattice_matrix, mommat=mommat)
                    self._other_coefficients[spin][label] = fite.fitde3D(
                        data, self._equivalences
                    )
            log_time_taken(t0)

    def get_amset_data(
        self,
        energy_cutoff: Optional[float] = None,
        scissor: float = None,
        bandgap: float = None,
        symprec: float = defaults["symprec"],
        nworkers: int = defaults["nworkers"],
    ) -> AmsetData:
        """Gets an AmsetData object using the interpolated bands.

        Note, the interpolation mesh is determined using by
        ``interpolate_factor`` option in the ``Inteprolater`` constructor.

        This method is much faster than the ``get_energies`` function but
        doesn't provide as much flexibility.

        The degree of parallelization is controlled by the ``nworkers`` option.

        Args:
            energy_cutoff: The energy cut-off to determine which bands are
                included in the interpolation. If the energy of a band falls
                within the cut-off at any k-point it will be included. For
                metals the range is defined as the Fermi level ± energy_cutoff.
                For gapped materials, the energy range is from the VBM -
                energy_cutoff to the CBM + energy_cutoff.
            scissor: The amount by which the band gap is scissored. Cannot
                be used in conjunction with the ``bandgap`` option. Has no
                effect for metallic systems.
            bandgap: Automatically adjust the band gap to this value. Cannot
                be used in conjunction with the ``scissor`` option. Has no
                effect for metallic systems.
            symprec: The symmetry tolerance used when determining the symmetry
                inequivalent k-points on which to interpolate.
            nworkers: The number of processors used to perform the
                interpolation. If set to ``-1``, the number of workers will
                be set to the number of CPU cores.

        Returns:
            The electronic structure (including energies, velocities, density of
            states and k-point information) as an AmsetData object.
        """
        is_metal = self._band_structure.is_metal()

        if is_metal and (bandgap or scissor):
            raise ValueError(
                "{} option set but system is metallic".format(
                    "bandgap" if bandgap else "scissor"
                )
            )

        nworkers = multiprocessing.cpu_count() if nworkers == -1 else nworkers

        logger.info("Interpolation parameters:")
        iinfo = [
            "k-point mesh: {}".format("x".join(map(str, self.interpolation_mesh))),
            "energy cutoff: {} eV".format(energy_cutoff),
        ]
        log_list(iinfo)

        ibands = get_ibands(energy_cutoff, self._band_structure)
        new_vb_idx = get_vb_idx(energy_cutoff, self._band_structure)

        energies = {}
        vvelocities = {}
        velocities = {}
        forgotten_electrons = 0
        for spin in self._spins:
            spin_ibands = ibands[spin]
            min_b = spin_ibands.min() + 1
            max_b = spin_ibands.max() + 1
            info = "Interpolating {} bands {}-{}".format(spin_name[spin], min_b, max_b)
            logger.info(info)

            # these are bands beneath the Fermi level that are dropped
            forgotten_electrons += min_b - 1

            t0 = time.perf_counter()
            energies[spin], vvelocities[spin], _, velocities[spin] = get_bands_fft(
                self._equivalences,
                self._coefficients[spin][spin_ibands],
                self._lattice_matrix,
                return_effective_mass=False,
                nworkers=nworkers,
            )
            log_time_taken(t0)

        if not self._soc and len(self._spins) == 1:
            forgotten_electrons *= 2
        nelectrons = self._num_electrons - forgotten_electrons

        if is_metal:
            efermi = self._band_structure.efermi * ev_to_hartree

        else:
            energies = _shift_energies(
                energies, new_vb_idx, scissor=scissor, bandgap=bandgap
            )

            # if material is semiconducting, set Fermi level to middle of gap
            efermi = _get_efermi(energies, new_vb_idx)

        logger.info("Generating tetrahedron mesh vertices")
        t0 = time.perf_counter()
        # get the actual k-points used in the BoltzTraP2 interpolation
        # unfortunately, BoltzTraP2 doesn't expose this information so we
        # have to get it ourselves
        (
            ir_kpts,
            _,
            full_kpts,
            ir_kpts_idx,
            ir_to_full_idx,
            tetrahedra,
            *ir_tetrahedra_info,
        ) = get_kpoints_tetrahedral(
            self.interpolation_mesh,
            self._band_structure.structure,
            symprec=symprec,
            time_reversal_symmetry=not self._soc,
        )
        log_time_taken(t0)

        energies, vvelocities, velocities = sort_amset_results(
            full_kpts, energies, vvelocities, velocities
        )
        atomic_structure = get_atomic_structure(self._band_structure.structure)

        return AmsetData(
            atomic_structure,
            energies,
            vvelocities,
            velocities,
            self.interpolation_mesh,
            full_kpts,
            ir_kpts_idx,
            ir_to_full_idx,
            tetrahedra,
            ir_tetrahedra_info,
            efermi,
            nelectrons,
            is_metal,
            self._soc,
            vb_idx=new_vb_idx,
        )

    def get_energies(
        self,
        kpoints: Union[np.ndarray, List],
        energy_cutoff: Optional[float] = None,
        scissor: float = None,
        bandgap: float = None,
        return_velocity: bool = False,
        return_curvature: bool = False,
        return_other_properties: bool = False,
        coords_are_cartesian: bool = False,
        atomic_units: bool = False,
        symprec: Optional[float] = defaults["symprec"],
        return_efermi: bool = False,
        return_vb_idx: bool = False,
    ) -> Union[Dict[Spin, np.ndarray], Tuple[Dict[Spin, np.ndarray], ...]]:
        """Gets the interpolated energies for multiple k-points in a band.

        Note, the accuracy of the interpolation is dependant on the
        ``interpolate_factor`` used to initialize the Interpolater.

        Args:
            kpoints: The k-point coordinates.
            energy_cutoff: The energy cut-off to determine which bands are
                included in the interpolation. If the energy of a band falls
                within the cut-off at any k-point it will be included. For
                metals the range is defined as the Fermi level ± energy_cutoff.
                For gapped materials, the energy range is from the VBM -
                energy_cutoff to the CBM + energy_cutoff.
            scissor: The amount by which the band gap is scissored. Cannot
                be used in conjunction with the ``bandgap`` option. Has no
                effect for metallic systems.
            bandgap: Automatically adjust the band gap to this value. Cannot
                be used in conjunction with the ``scissor`` option. Has no
                effect for metallic systems.
            return_velocity: Whether to return the band velocities.
            return_curvature: Whether to return the band curvature (inverse effective
                mass).
            return_other_properties: Whether to return the interpolated results
                for the ``other_properties`` data.
            coords_are_cartesian: Whether the kpoints are in cartesian or
                fractional coordinates.
            atomic_units: Return the energies, velocities, and effective_massses
                in atomic units. If False, energies will be in eV, velocities in
                cm/s, and curvature in units of 1 / electron rest mass (1/m0).
            symprec: Symmetry precision. If set, symmetry will be used to
                reduce the nummber of calculated k-points and velocities.
            return_efermi: Whether to return the Fermi level with the unit
                determined by ``atomic_units``. If the system is semiconducting
                the Fermi level will be given in the middle of the band gap.
            return_vb_idx: Whether to return the index of the highest valence band
                in the interpolated bands. Will be returned as a dictionary of
                ``{spin: vb_idx}``.

        Returns:
            The band energies as dictionary of::

                {spin: energies}

            If ``return_velocity``, ``curvature`` or ``return_other_properties`` a tuple
            is returned, formatted as::

                (energies, Optional[velocities], Optional[curvature],
                 Optional[other_properties])

            The velocities and effective masses are given as the 1x3 trace and
            full 3x3 tensor, respectively (along cartesian directions). The
            projections are summed for each orbital type (s, p, d) across all
            atoms, and are given as::

                {spin: {orbital: projections}}
        """
        if self._band_structure.is_metal() and (bandgap or scissor):
            raise ValueError(
                "{} option set but system is metallic".format(
                    "bandgap" if bandgap else "scissor"
                )
            )

        # only calculate the energies for the bands within the energy cutoff
        lattice = self._band_structure.structure.lattice
        kpoints = np.asarray(kpoints)
        nkpoints = len(kpoints)

        if coords_are_cartesian:
            kpoints = lattice.reciprocal_lattice.get_fractional_coords(kpoints)

        if symprec:
            logger.info("Reducing # k-points using symmetry")
            (
                kpoints,
                weights,
                ir_kpoints_idx,
                ir_to_full_idx,
                _,
                rot_mapping,
            ) = get_symmetry_equivalent_kpoints(
                self._band_structure.structure,
                kpoints,
                symprec=symprec,
                return_inverse=True,
                time_reversal_symmetry=not self._soc,
            )
            k_info = [
                "# original k-points: {}".format(nkpoints),
                "# reduced k-points {}".format(len(kpoints)),
            ]
            log_list(k_info)

        ibands = get_ibands(energy_cutoff, self._band_structure)
        new_vb_idx = get_vb_idx(energy_cutoff, self._band_structure)

        energies = {}
        velocities = {}
        curvature = {}
        other_properties = defaultdict(dict)
        for spin in self._spins:
            spin_ibands = ibands[spin]
            min_b = spin_ibands.min() + 1
            max_b = spin_ibands.max() + 1
            info = "Interpolating {} bands {}-{}".format(spin_name[spin], min_b, max_b)
            logger.info(info)

            t0 = time.perf_counter()
            fitted = fite.getBands(
                kpoints,
                self._equivalences,
                self._lattice_matrix,
                self._coefficients[spin][spin_ibands],
                curvature=return_curvature,
            )
            log_time_taken(t0)

            energies[spin] = fitted[0]
            velocities[spin] = fitted[1]

            if return_curvature:
                # make curvature have the shape ((nbands, nkpoints, 3, 3)
                curvature[spin] = fitted[2]
                curvature[spin] = curvature[spin].transpose((2, 3, 0, 1))

            if return_other_properties:
                logger.info("Interpolating {} properties".format(spin_name[spin]))

                t0 = time.perf_counter()
                for label, coeffs in self._other_coefficients[spin].items():
                    other_properties[spin][label], _ = fite.getBands(
                        kpoints,
                        self._equivalences,
                        self._lattice_matrix,
                        coeffs[spin_ibands],
                        curvature=False,
                    )

                log_time_taken(t0)

            if not atomic_units:
                energies[spin] = energies[spin] * hartree_to_ev
                velocities[spin] = _convert_velocities(velocities[spin], lattice.matrix)

        if symprec:
            energies, velocities, curvature, other_properties = symmetrize_results(
                energies,
                velocities,
                curvature,
                other_properties,
                ir_to_full_idx,
                rot_mapping,
                self._band_structure.structure.lattice.reciprocal_lattice.matrix,
            )

        if not self._band_structure.is_metal():
            energies = _shift_energies(
                energies, new_vb_idx, scissor=scissor, bandgap=bandgap
            )

        to_return = [energies]

        if return_velocity:
            to_return.append(velocities)

        if return_curvature:
            to_return.append(curvature)

        if return_other_properties:
            to_return.append(other_properties)

        if return_efermi:
            if self._band_structure.is_metal():
                efermi = self._band_structure.efermi
                if atomic_units:
                    efermi *= ev_to_hartree
            else:
                # if semiconducting, set Fermi level to middle of gap
                efermi = _get_efermi(energies, new_vb_idx)

            to_return.append(efermi)

        if return_vb_idx:
            to_return.append(new_vb_idx)

        if len(to_return) == 1:
            return to_return[0]
        else:
            return tuple(to_return)

    def get_dos(
        self,
        kpoint_mesh: Union[float, int, List[int]],
        energy_cutoff: Optional[float] = None,
        scissor: Optional[float] = None,
        bandgap: Optional[float] = None,
        estep: float = defaults["dos_estep"],
        symprec: float = defaults["symprec"],
        atomic_units: bool = False,
    ) -> Union[Dos, FermiDos]:
        """Calculates the density of states using the interpolated bands.

        Args:
            kpoint_mesh: The k-point mesh as a 1x3 array. E.g.,``[6, 6, 6]``.
                Alternatively, if a single value is provided this will be
                treated as a reciprocal density and the k-point mesh dimensions
                generated automatically.
            energy_cutoff: The energy cut-off to determine which bands are
                included in the interpolation. If the energy of a band falls
                within the cut-off at any k-point it will be included. For
                metals the range is defined as the Fermi level ± energy_cutoff.
                For gapped materials, the energy range is from the VBM -
                energy_cutoff to the CBM + energy_cutoff.
            scissor: The amount by which the band gap is scissored. Cannot
                be used in conjunction with the ``bandgap`` option. Has no
                effect for metallic systems.
            bandgap: Automatically adjust the band gap to this value. Cannot
                be used in conjunction with the ``scissor`` option. Has no
                effect for metallic systems.
            estep: The energy step, where smaller numbers give more
                accuracy but are more expensive.
            symprec: The symmetry tolerance used when determining the symmetry
                inequivalent k-points on which to interpolate.
            atomic_units: Whether to return the DOS in atomic units. If False, the
                unit of energy will be eV.

        Returns:
            The density of states.
        """
        if isinstance(kpoint_mesh, numeric_types):
            logger.info("DOS k-point length cutoff: {}".format(kpoint_mesh))
        else:
            str_mesh = "x".join(map(str, kpoint_mesh))
            logger.info("DOS k-point mesh: {}".format(str_mesh))

        structure = self._band_structure.structure
        tri = not self._soc
        (
            ir_kpts,
            _,
            full_kpts,
            ir_kpts_idx,
            ir_to_full_idx,
            tetrahedra,
            *ir_tetrahedra_info,
        ) = get_kpoints_tetrahedral(
            kpoint_mesh, structure, symprec=symprec, time_reversal_symmetry=tri
        )

        energies, efermi, vb_idx = self.get_energies(
            ir_kpts,
            scissor=scissor,
            bandgap=bandgap,
            energy_cutoff=energy_cutoff,
            atomic_units=atomic_units,
            return_efermi=True,
            return_vb_idx=True,
        )

        if not self._band_structure.is_metal():
            # if not a metal, set the Fermi level to the top of the valence band.
            efermi = np.max([np.max(e[vb_idx[spin]]) for spin, e in energies.items()])

        full_energies = {s: e[:, ir_to_full_idx] for s, e in energies.items()}
        tetrahedral_band_structure = TetrahedralBandStructure.from_data(
            full_energies,
            full_kpts,
            tetrahedra,
            structure,
            ir_kpts_idx,
            ir_to_full_idx,
            *ir_tetrahedra_info
        )

        emin = np.min([np.min(spin_eners) for spin_eners in energies.values()])
        emax = np.max([np.max(spin_eners) for spin_eners in energies.values()])
        epoints = int(round((emax - emin) / estep))
        energies = np.linspace(emin, emax, epoints)

        _, dos = tetrahedral_band_structure.get_density_of_states(energies)

        return FermiDos(efermi, energies, dos, structure, atomic_units=atomic_units)

    def get_line_mode_band_structure(
        self,
        line_density: int = 50,
        kpath: Optional[Kpath] = None,
        energy_cutoff: Optional[float] = None,
        scissor: Optional[float] = None,
        bandgap: Optional[float] = None,
        symprec: Optional[float] = defaults["symprec"],
        return_other_properties: bool = False,
    ) -> Union[
        BandStructureSymmLine,
        Tuple[BandStructureSymmLine, Dict[Spin, Dict[str, np.ndarray]]],
    ]:
        """Gets the interpolated band structure along high symmetry directions.

        Args:
            line_density: The maximum number of k-points between each two
                consecutive high-symmetry k-points
            energy_cutoff: The energy cut-off to determine which bands are
                included in the interpolation. If the energy of a band falls
                within the cut-off at any k-point it will be included. For
                metals the range is defined as the Fermi level ± energy_cutoff.
                For gapped materials, the energy range is from the VBM -
                energy_cutoff to the CBM + energy_cutoff.
            scissor: The amount by which the band gap is scissored. Cannot
                be used in conjunction with the ``bandgap`` option. Has no
                effect for metallic systems.
            bandgap: Automatically adjust the band gap to this value. Cannot
                be used in conjunction with the ``scissor`` option. Has no
                effect for metallic systems.
            symprec: The symmetry tolerance used to determine the space group
                and high-symmetry path.
            return_other_properties: Whether to include the interpolated
                other_properties data for each k-point along the band structure path.

        Returns:
            The line mode band structure.
        """
        if not kpath:
            kpath = PymatgenKpath(self._band_structure.structure, symprec=symprec)

        kpoints, labels = kpath.get_kpoints(line_density=line_density, cart_coords=True)
        labels_dict = {
            label: kpoint for kpoint, label in zip(kpoints, labels) if label != ""
        }

        energies = self.get_energies(
            kpoints,
            scissor=scissor,
            bandgap=bandgap,
            atomic_units=False,
            energy_cutoff=energy_cutoff,
            coords_are_cartesian=True,
            return_other_properties=return_other_properties,
            symprec=symprec,
        )

        if return_other_properties:
            energies, other_properties = energies

        bs = BandStructureSymmLine(
            kpoints,
            energies,
            self._band_structure.structure.lattice,
            self._band_structure.efermi,
            labels_dict,
            coords_are_cartesian=True,
        )
        if return_other_properties:
            return bs, other_properties
        else:
            return bs


class DFTData(object):
    """DFTData object used for BoltzTraP2 interpolation.

    Note that the units used by BoltzTraP are different to those used by VASP.

    Args:
        kpoints: The k-points in fractional coordinates.
        energies: The band energies in Hartree, formatted as (nbands, nkpoints).
        lattice_matrix: The lattice matrix in Bohr^3.
        mommat: The band structure derivatives.
    """

    def __init__(
        self,
        kpoints: np.ndarray,
        energies: np.ndarray,
        lattice_matrix: np.ndarray,
        mommat: Optional[np.ndarray] = None,
    ):
        self.kpoints = kpoints
        self.ebands = energies
        self.lattice_matrix = lattice_matrix
        self.volume = np.abs(np.linalg.det(self.lattice_matrix))
        self.mommat = mommat

    def get_lattvec(self) -> np.ndarray:
        """Get the lattice matrix. This method is required by BoltzTraP2."""
        return self.lattice_matrix


def _shift_energies(
    energies: Dict[Spin, np.ndarray],
    vb_idx: Dict[Spin, int],
    scissor: Optional[float] = None,
    bandgap: Optional[float] = None,
) -> Union[Dict[Spin, np.ndarray], Tuple[Dict[Spin, np.ndarray], float]]:
    """Shift the band energies based on the scissor or bandgap parameter.

    Args:
        energies: The band energies in Hartree, given for each Spin channel.
        vb_idx: The band index of the valence band maximum in the energies
            array, given for each Spin channel.
        scissor: The amount by which the band gap is scissored. Cannot
            be used in conjunction with the ``bandgap`` option. Has no
            effect for metallic systems.
        bandgap: Automatically adjust the band gap to this value. Cannot
            be used in conjunction with the ``scissor`` option. Has no
            effect for metallic systems.

    Returns:
        The energies, shifted according to ``scissor`` or ``bandgap``. If
        return_scissor is True, a tuple of (energies, scissor) is returned.
    """

    if scissor and bandgap:
        raise ValueError("scissor and bandgap cannot be set simultaneously")

    cb_idx = {s: v + 1 for s, v in vb_idx.items()}

    if bandgap:
        interp_bandgap = (
            min([energies[s][cb_idx[s] :].min() for s in energies])
            - max([energies[s][: cb_idx[s]].max() for s in energies])
        ) * hartree_to_ev

        scissor = bandgap - interp_bandgap
        logger.info(
            "Bandgap set to {:.3f} eV, automatically scissoring by "
            "{:.3f} eV".format(bandgap, scissor)
        )

    if scissor:
        scissor *= ev_to_hartree
        for spin in energies:
            energies[spin][: cb_idx[spin]] -= scissor / 2
            energies[spin][cb_idx[spin] :] += scissor / 2

    return energies


def _convert_velocities(
    velocities: np.ndarray, lattice_matrix: np.ndarray
) -> np.ndarray:
    """Convert velocities from atomic units to cm/s.

    Args:
        velocities: The velocities in atomic units.
        lattice_matrix: The lattice matrix in Angstrom.

    Returns:
        The velocities in cm/s.
    """
    velocities *= bohr_to_cm / au_to_s
    return velocities


def _get_efermi(energies: Dict[Spin, np.ndarray], vb_idx: Dict[Spin, int]) -> float:
    e_vbm = max(
        [np.max(energies[spin][: vb_idx[spin] + 1]) for spin in energies.keys()]
    )
    e_cbm = min(
        [np.min(energies[spin][vb_idx[spin] + 1 :]) for spin in energies.keys()]
    )
    return (e_vbm + e_cbm) / 2


def symmetrize_results(
    energies,
    velocities,
    curvature,
    other_properties,
    ir_to_full_idx,
    rotation_matrices,
    reciprocal_lattice,
):
    similarity_matrices = np.array(
        [similarity_transformation(reciprocal_lattice, r) for r in rotation_matrices]
    )
    inv_similarity_matrices = np.array([np.linalg.inv(s) for s in similarity_matrices])

    for spin in energies:
        energies[spin] = energies[spin][:, ir_to_full_idx]
        velocities[spin] = rotate_velocities(
            velocities[spin][..., ir_to_full_idx], similarity_matrices
        )

        if curvature:
            curvature[spin] = rotate_curvature(
                curvature[spin][:, ir_to_full_idx, ...],
                similarity_matrices,
                inv_similarity_matrices,
            )

        if other_properties:
            for label, prop in other_properties[spin].items():
                other_properties[spin][label] = prop[:, ir_to_full_idx]

    return energies, velocities, curvature, other_properties


def rotate_velocities(velocities, similarity_matrices):
    # apply rotation matrices to the velocities at the symmetry
    # reduced k-points, to get the velocities for the full
    # original mesh (this is just the dot product of the velocity
    # and appropriate rotation matrix. The weird ordering of the
    # indices is because the velocities has the shape
    # (3, nbands, nkpoints)
    return np.einsum("jkl,kij->lij", similarity_matrices, velocities)


def rotate_curvature(curvature, similarity_matrices, inv_similarity_matrices):
    rot_curvature = np.empty(curvature.shape)

    for b_idx, k_idx in np.ndindex(curvature.shape[:2]):
        inner = np.dot(curvature[b_idx, k_idx], similarity_matrices[k_idx])
        rot_curvature[b_idx, k_idx] = np.dot(inv_similarity_matrices[k_idx], inner)

    return rot_curvature


def sort_amset_results(kpoints, energies, vvelocities, velocities):
    # BoltzTraP2 and spglib give k-points in different orders. We need to use the
    # spglib ordering to make the tetrahedron method work, so get the indices
    # that will sort from BoltzTraP2 order to spglib order
    sort_idx = sort_boltztrap_to_spglib(kpoints)

    energies = {s: ener[:, sort_idx] for s, ener in energies.items()}
    vvelocities = {s: vv[:, :, :, sort_idx] for s, vv in vvelocities.items()}
    velocities = {s: v[:, :, sort_idx] for s, v in velocities.items()}
    return energies, vvelocities, velocities



import datetime
import logging
import math
import sys
import textwrap
import time
from pathlib import Path
from typing import Union

from amset.constants import output_width

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


def initialize_amset_logger(
    directory: Union[str, Path] = ".",
    filename: Union[str, Path, bool] = "amset.log",
    level: int = logging.INFO,
    print_log: bool = True,
) -> logging.Logger:
    """Initialize the default logger with stdout and file handlers.

    Args:
        directory: Path to the folder where the log file will be written.
        filename: The log filename. If False, no log will be written.
        level: The log level.
        print_log: Whether to print the log to the screen.

    Returns:
        A logging instance with customized formatter and handlers.
    """
    log = logging.getLogger("amset")
    log.setLevel(level)
    log.handlers = []  # reset logging handlers if they already exist

    screen_formatter = WrappingFormatter(fmt="%(message)s")
    file_formatter = WrappingFormatter(fmt="%(message)s", simple_ascii=True)

    if filename is not False:
        handler = logging.FileHandler(Path(directory) / filename, mode="w")
        handler.setFormatter(file_formatter)
        log.addHandler(handler)

    if print_log:
        screen_handler = logging.StreamHandler(stream=sys.stdout)
        screen_handler.setFormatter(screen_formatter)
        log.addHandler(screen_handler)

    def handle_exception(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        now = datetime.datetime.now()
        exit_msg = "amset exiting on {} at {}".format(
            now.strftime("%d %b %Y"), now.strftime("%H:%M")
        )

        log.error(
            "\n  ERROR: {}".format(exit_msg),
            exc_info=(exc_type, exc_value, exc_traceback),
        )

    sys.excepthook = handle_exception

    return log


class WrappingFormatter(logging.Formatter):
    def __init__(
        self, fmt=None, datefmt=None, style="%", width=output_width, simple_ascii=False
    ):
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)
        self.simple_ascii = simple_ascii
        self.wrapper = textwrap.TextWrapper(
            width=width,
            subsequent_indent="  ",
            replace_whitespace=True,
            drop_whitespace=False,
        )

    def format(self, record):
        text = super().format(record)
        if "└" in text or "├" in text:
            # don't have blank time when reporting list
            text = "  " + text
        else:
            text = "\n" + "\n".join(
                [self.wrapper.fill("  " + s) for s in text.splitlines()]
            )

        if self.simple_ascii:
            return self.make_simple_ascii(text)

        return text

    @staticmethod
    def make_simple_ascii(text):
        replacements = {
            "├──": "-",
            "│": " ",
            "└──": "-",
            fancy_logo: simple_logo,
            "ᵢᵢ": "_ii",
            "ħω": "hbar.omega",
            "cm²/Vs": "cm^2/Vs",
            "β²": "beta^2",
            "a₀⁻²": "a_0^-2",
            "cm⁻³": "cm^-3",
            "–": "-",
            "₀": "0",
            "₁": "1",
            "₂": "2",
            "₃": "3",
            "₄": "4",
            "₅": "5",
            "₆": "6",
            "₇": "7",
            "₈": "8",
            "₉": "8",
            "\u0305": "-",
            "π": "pi",
            "ħ": "hbar",
            "ω": "omega",
            "α": "alpha",
            "β": "beta",
            "γ": "gamma",
            "°": "deg",
        }

        for initial, final in replacements.items():
            text = text.replace(initial, final)
        return text


def log_time_taken(t0: float):
    logger.info("  └── time: {:.4f} s".format(time.perf_counter() - t0))


def log_banner(text):
    width = output_width - 2
    nstars = (width - (len(text) + 2)) / 2
    logger.info(
        "\n{} {} {}".format("~" * math.ceil(nstars), text, "~" * math.floor(nstars))
    )


def log_list(list_strings, prefix="  ", level=logging.INFO):
    for i, text in enumerate(list_strings):
        if i == len(list_strings) - 1:
            pipe = "└"
        else:
            pipe = "├"
        logger.log(level, "{}{}── {}".format(prefix, pipe, text))


fancy_logo = """                 █████╗ ███╗   ███╗███████╗███████╗████████╗
                ██╔══██╗████╗ ████║██╔════╝██╔════╝╚══██╔══╝
                ███████║██╔████╔██║███████╗█████╗     ██║
                ██╔══██║██║╚██╔╝██║╚════██║██╔══╝     ██║
                ██║  ██║██║ ╚═╝ ██║███████║███████╗   ██║
                ╚═╝  ╚═╝╚═╝     ╚═╝╚══════╝╚══════╝   ╚═╝
"""

simple_logo = r"""         /$$$$$$  /$$      /$$  /$$$$$$  /$$$$$$$$ /$$$$$$$$
        /$$__  $$| $$$    /$$$ /$$__  $$| $$_____/|__  $$__/
       | $$  \ $$| $$$$  /$$$$| $$  \__/| $$         | $$
       | $$$$$$$$| $$ $$/$$ $$|  $$$$$$ | $$$$$      | $$
       | $$__  $$| $$  $$$| $$ \____  $$| $$__/      | $$
       | $$  | $$| $$\  $ | $$ /$$  \ $$| $$         | $$
       | $$  | $$| $$ \/  | $$|  $$$$$$/| $$$$$$$$   | $$
       |__/  |__/|__/     |__/ \______/ |________/   |__/
"""



import logging
import time

import numpy as np

from amset.electronic_structure.symmetry import similarity_transformation
from amset.log import log_time_taken
from amset.util import get_progress_bar

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

_voigt_idxs = ([0, 1, 2, 1, 0, 0], [0, 1, 2, 2, 2, 1])
_fmt_str = (
    "[{0:{space}.{dp}f} {1:{space}.{dp}f} {2:{space}.{dp}f} {3:{space}.{dp}f} "
    "{4:{space}.{dp}f} {5:{space}.{dp}f}]"
)

logger = logging.getLogger(__name__)


def get_formatted_tensors(tensors):
    min_n = np.inf
    for tensor in tensors:
        tensor = tensor.round(5)
        frac_t = tensor - np.round(tensor)
        min_n = min(np.min(np.abs(frac_t[np.nonzero(frac_t)])), min_n)

    # get the number of decimal places we will need
    dp = np.ceil(np.log10(1 / min_n)).astype(int)
    space = dp + 3

    formatted_tensors = []
    for tensor in tensors:
        voigt = np.array(tensor)[_voigt_idxs].round(5)
        voigt[voigt == -0.0] = 0
        formatted_tensors.append(_fmt_str.format(*voigt, space=space, dp=dp))

    return formatted_tensors


def desymmetrize_deformation_potentials(
    deformation_potentials, structure, rotations, op_mapping, kp_mapping, pbar=True
):
    logger.info("Desymmetrizing deformation potentials")
    t0 = time.perf_counter()

    rlat = structure.lattice.reciprocal_lattice.matrix
    sims = np.array([similarity_transformation(rlat, r) for r in rotations])
    inv_sims = np.array([np.linalg.inv(s) for s in sims])

    sims = sims[op_mapping]
    inv_sims = inv_sims[op_mapping]

    all_deformation_potentials = {}
    for spin, spin_deformation_potentials in deformation_potentials.items():
        all_deformation_potentials[spin] = np.zeros(
            (len(spin_deformation_potentials), len(sims), 3, 3)
        )

        state_idxs = list(np.ndindex((len(spin_deformation_potentials), len(sims))))
        if pbar:
            state_idxs = get_progress_bar(state_idxs, desc="progress")

        for b_idx, k_idx in state_idxs:
            map_idx = kp_mapping[k_idx]

            sim = sims[k_idx]
            inv_sim = inv_sims[k_idx]

            inner = np.dot(sim, spin_deformation_potentials[b_idx, map_idx])
            rot_deform = np.abs(np.dot(inner, inv_sim))

            # inner = np.dot(spin_deformation_potentials[b_idx, map_idx], inv_sim)
            # rot_deform = np.abs(np.dot(sim, inner))
            # inner = np.dot(spin_deformation_potentials[b_idx, map_idx], sim)
            # rot_deform = np.abs(np.dot(inv_sim, inner))

            all_deformation_potentials[spin][b_idx, k_idx] = rot_deform

    log_time_taken(t0)
    return all_deformation_potentials



import numpy as np
from sumo.plotting import pretty_plot, styled_plot

from amset.constants import bohr_to_m, bohr_to_nm, boltzmann_au, s_to_au
from amset.electronic_structure.fd import dfdde
from amset.plot import BaseMeshPlotter, amset_base_style

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

_y_labels = {"conductivity": r"Cumulative conductivity (\%)"}

_x_labels = {
    "mean free path": "Mean free path (nm)",
    "group velocity": "Group velocity (m/s)",
    "scattering rate": r"Scattering rate (s$^{-1}$)",
}

_conversions = {
    "mean free path": bohr_to_nm,
    "group velocity": bohr_to_m,
    "scattering rate": 1,
}


class CumulativePlotter(BaseMeshPlotter):
    @styled_plot(amset_base_style)
    def get_plot(
        self,
        n_idx,
        t_idx,
        x_property="mean free path",
        y_property="conductivity",
        height=6,
        width=6,
        xlabel=None,
        ylabel=None,
        xlim=None,
        ylim=None,
        logx=False,
        plt=None,
        style=None,
        no_base_style=False,
        fonts=None,
    ):
        x_values, y_values = self.get_plot_data(n_idx, t_idx, x_property, y_property)

        plt = pretty_plot(width=width, height=height, plt=plt)
        ax = plt.gca()
        ax.plot(x_values, y_values)

        xlabel = xlabel if xlabel else _x_labels[x_property.lower()]
        ylabel = ylabel if ylabel else _y_labels[y_property.lower()]
        ax.set(xlabel=xlabel, ylabel=ylabel, xlim=xlim, ylim=ylim)

        if logx:
            ax.semilogx()

    def get_plot_data(
        self, n_idx, t_idx, x_property="mean free path", y_property="conductivity"
    ):
        x_values = self._get_x_values(n_idx, t_idx, x_property)
        y_values = self._get_y_values(n_idx, t_idx, y_property)
        x_values, y_values = _get_cummulative_sum(x_values, y_values)

        x_values *= _conversions[x_property.lower()]
        y_values *= 100  # to percent

        return x_values, y_values

    def _get_y_values(self, n_idx, t_idx, y_property):
        if y_property.lower() == "conductivity":
            y_values = self._get_conductivity(n_idx, t_idx)
        else:
            raise ValueError("unknown y_property: {}".format(y_property))

        return np.ravel(y_values)

    def _get_x_values(self, n_idx, t_idx, x_property):
        if x_property.lower() == "mean free path":
            x_values = self._get_mean_free_path(n_idx, t_idx)
        elif x_property.lower() == "group velocity":
            x_values = self._get_group_velocity()
        elif x_property.lower() == "scattering rate":
            x_values = self._get_scattering_rates(n_idx, t_idx)
        else:
            raise ValueError("unknown x_property: {}".format(x_property))

        return np.ravel(x_values)

    def _get_conductivity(self, n_idx, t_idx):
        velocities = self._get_group_velocity()
        lifetimes = 1 / self._get_scattering_rates(n_idx, t_idx)

        energies = self._get_energies()
        _, weights = np.unique(self.ir_to_full_kpoint_mapping, return_counts=True)

        ef = self.fermi_levels[n_idx, t_idx]
        temp = self.temperatures[t_idx]

        dfde = -dfdde(energies, ef, temp * boltzmann_au)
        nkpoints = len(self.kpoints)

        integrand = velocities ** 2 * lifetimes * dfde * weights[None, :] / nkpoints
        conductivity = np.sum(integrand)

        return integrand / conductivity

    def _get_mean_free_path(self, n_idx, t_idx):
        # mean free path in bohr
        group_velocity = self._get_group_velocity()
        scattering_rate = self._get_scattering_rates(n_idx, t_idx)
        return group_velocity / scattering_rate

    def _get_group_velocity(self):
        # mean free path in bohr / s
        velocities = {s: np.linalg.norm(v, axis=2) for s, v in self.velocities.items()}
        all_velocities = np.concatenate([v for v in velocities.values()])
        all_velocities[all_velocities < 0.005] = 0.005  # handle very small velocities
        return all_velocities * s_to_au

    def _get_scattering_rates(self, n_idx, t_idx):
        rates = self.scattering_rates.values()
        return np.concatenate([np.sum(r[:, n_idx, t_idx], axis=0) for r in rates])

    def _get_energies(self):
        return np.concatenate([e for e in self.energies.values()])


def _get_cummulative_sum(x_values, y_values):
    sort_idx = np.argsort(x_values)
    y_values = y_values[sort_idx]
    x_values = x_values[sort_idx]
    cumsum = np.cumsum(y_values)
    return x_values, cumsum



import warnings

from amset.core.run import Runner
from amset.plot.rates import RatesPlotter

warnings.simplefilter("ignore")

settings = {
    # general settings
    "scattering_type": ["IMP", "ADP"],
    "doping": [-1.99e14, -2.20e15, -1.72e16, -1.86e17, -1.46e18, -4.39e18],
    "temperatures": [300],
    "bandgap": 1.14,

    # electronic_structure settings
    "interpolation_factor": 50,

    # materials properties
    "deformation_potential": "deformation.h5",
    "elastic_constant": [[144,  53,  53,  0,  0,  0],
                         [53,  144,  53,  0,  0,  0],
                         [53,   53, 144,  0,  0,  0],
                         [0,     0,   0, 75,  0,  0],
                         [0,     0,   0,  0, 75,  0],
                         [0,     0,   0,  0,  0, 75]],
    "static_dielectric": [[11.7, 0, 0], [0, 11.7, 0], [0, 0, 11.7]],
    "high_frequency_dielectric": [[11.7, 0, 0], [0, 11.7, 0], [0, 0, 11.7]],

    # performance settings
    "write_mesh": True,
}

if __name__ == "__main__":
    runner = Runner.from_vasprun("vasprun.xml.gz", settings)
    amset_data = runner.run()

    plotter = RatesPlotter(amset_data)
    plt = plotter.get_plot()
    plt.savefig("Si_rates.png", bbox_inches="tight", dpi=400)



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple

import numpy as np

from amset.constants import nm_to_bohr, s_to_au
from amset.core.data import AmsetData

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class AbstractBasicScattering(ABC):

    name: str
    required_properties: Tuple[str]

    def __init__(self, materials_properties: Dict[str, Any], amset_data: AmsetData):
        self.properties = {p: materials_properties[p] for p in self.required_properties}
        self.doping = amset_data.doping
        self.temperatures = amset_data.temperatures
        self.nbands = {s: len(amset_data.energies[s]) for s in amset_data.spins}
        self.spins = amset_data.spins

    @property
    @abstractmethod
    def rates(self):
        pass


class ConstantRelaxationTime(AbstractBasicScattering):

    name = "CRT"
    required_properties = ("constant_relaxation_time",)

    def __init__(self, materials_properties: Dict[str, Any], amset_data: AmsetData):
        super().__init__(materials_properties, amset_data)
        rate = 1 / self.properties["constant_relaxation_time"]
        shape = {
            s: amset_data.fermi_levels.shape + amset_data.energies[s].shape
            for s in self.spins
        }
        self._rates = {s: np.full(shape[s], rate) for s in self.spins}

    @property
    def rates(self):
        # need to return rates with shape (nspins, ndops, ntemps, nbands, nkpoints)
        return self._rates


class MeanFreePathScattering(AbstractBasicScattering):

    name = "MFP"
    required_properties = ("mean_free_path",)

    def __init__(self, materials_properties: Dict[str, Any], amset_data: AmsetData):
        super().__init__(materials_properties, amset_data)

        # convert mean free path from nm to bohr
        mfp = self.properties["mean_free_path"] * nm_to_bohr

        self._rates = {}
        ir_kpoints_idx = amset_data.ir_kpoints_idx
        for spin in self.spins:
            vvelocities = amset_data.velocities_product[spin][..., ir_kpoints_idx]
            v = np.sqrt(np.diagonal(vvelocities, axis1=1, axis2=2))
            v = np.linalg.norm(v, axis=2)
            v[v < 0.005] = 0.005  # handle very small velocities
            velocities = np.tile(v, (len(self.doping), len(self.temperatures), 1, 1))

            rates = velocities * s_to_au / mfp
            self._rates[spin] = rates[..., amset_data.ir_to_full_kpoint_mapping]

    @property
    def rates(self):
        # need to return rates with shape (nspins, ndops, ntemps, nbands, nkpoints)
        return self._rates



import abc
from pathlib import Path
from typing import Union

from monty.serialization import loadfn
from pkg_resources import resource_filename

from amset.core.data import AmsetData
from amset.io import load_mesh
from amset.util import cast_dict_ndarray

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

amset_base_style = resource_filename("amset.plot", "amset_base.mplstyle")
revtex_style = resource_filename("amset.plot", "revtex.mplstyle")


class BaseMeshPlotter(abc.ABC):
    def __init__(self, mesh_data: Union[str, Path, AmsetData, dict]):
        if isinstance(mesh_data, (str, Path)):
            if "h5" not in mesh_data:
                raise ValueError(
                    "mesh.h5 file needed for plot. Run AMSET with write_mesh=True"
                )
            mesh_data = load_mesh(mesh_data)
        elif isinstance(mesh_data, AmsetData):
            mesh_data = mesh_data.to_dict(include_mesh=True)["mesh"]
        elif "mesh" in mesh_data:
            mesh_data = mesh_data["mesh"]
        elif not isinstance(mesh_data, dict):
            raise ValueError("Unrecognised data format")

        self._data = cast_dict_ndarray(mesh_data)
        self.spins = list(self.energies.keys())

    def __getattr__(self, item):
        return self._data[item]


class BaseTransportPlotter(abc.ABC):
    def __init__(self, data: Union[str, Path, AmsetData, dict]):
        if isinstance(data, (str, Path)):
            # guess the mesh filename
            data = loadfn(data)
        elif isinstance(data, AmsetData):
            data = data.to_dict(include_mesh=True)
        elif isinstance(data, dict):
            data = data
        else:
            raise ValueError("Unrecognised data format")

        self._data = cast_dict_ndarray(data)
        self.spins = list(self.energies.keys())

    def __getattr__(self, item):
        return self._data[item]



from pathlib import Path

import h5py
import numpy as np
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.io.vasp.outputs import Outcar, Vasprun

from amset.constants import defaults, str_to_spin
from amset.electronic_structure.common import get_band_structure

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


def write_deformation_potentials(
    deformation_potentials, kpoints, structure, filename="deformation.h5"
):
    with h5py.File(filename, "w") as f:
        for spin, spin_deform in deformation_potentials.items():
            name = "deformation_potentials_{}".format(spin.name)
            f.create_dataset(name, data=spin_deform, compression="gzip")
        f["structure"] = np.string_(structure.to_json())
        f["kpoints"] = kpoints
    return filename


def load_deformation_potentials(filename):
    deformation_potentials = {}
    with h5py.File(filename, "r") as f:
        deform_keys = [k for k in list(f.keys()) if "deformation_potentials" in k]
        for key in deform_keys:
            spin = str_to_spin[key.split("_")[-1]]
            deformation_potentials[spin] = np.array(f[key])

        structure_str = np.string_(np.array(f["structure"])).decode()
        structure = Structure.from_str(structure_str, fmt="json")
        kpoints = np.array(f["kpoints"])

    return deformation_potentials, kpoints, structure


def write_deformed_poscars(deformed_structures, directory=None):
    if not directory:
        directory = "."
    directory = Path(directory)
    n_deformations = len(deformed_structures)
    n_digits = int(np.floor(np.log10(n_deformations)) + 2)
    for i, deformed_structure in enumerate(deformed_structures):
        # pad with leading zeros so the files are sorted correctly
        filename = "POSCAR-{0:0{1}}".format(i + 1, n_digits)
        deform_poscar = Poscar(deformed_structure)
        deform_poscar.write_file(directory / filename, significant_figures=16)


def parse_calculation(folder, zero_weighted_kpoints=defaults["zero_weighted_kpoints"]):
    vr = Vasprun(get_gzipped_file("vasprun.xml", folder))
    out = Outcar(get_gzipped_file("OUTCAR", folder))
    bs = get_band_structure(vr, zero_weighted=zero_weighted_kpoints)
    reference_level = get_reference_energy(bs, out)
    return {"reference": reference_level, "bandstructure": bs}


def get_gzipped_file(filename, folder):
    folder = Path(folder)
    gz_filename = filename + ".gz"
    if (folder / filename).exists():
        return folder / filename
    elif (folder / gz_filename).exists():
        return folder / gz_filename
    else:
        raise FileNotFoundError("Could not find {} file in {}".format(filename, folder))


def get_reference_energy(bandstructure, outcar):
    if bandstructure.is_metal():
        return bandstructure.efermi
    else:
        # read the average potential at atomic cores from the OUTCAR; note: if
        # ICORELEVEL = 1, these will be not be written and pot will be an empty list
        pot = outcar.read_avg_core_poten()

        if len(pot) > 0:
            return np.mean(pot[0])

        # read the core level eigenvalues from the OUTCAR; note: if ICORELEVEL
        # is not set these will be not be written and eigen will be a list
        # of empty dictionaries
        eigen = outcar.read_core_state_eigen()
        ref = [x["1s"][0] for x in eigen if "1s" in x]
        if len(ref) > 0:
            return np.mean(ref)
        else:
            raise ValueError(
                "OUTCAR does not contain avg electrostatic potential or eigenvalues"
            )



import logging

import numpy as np
from interpolation.splines import UCGrid, eval_linear
from interpolation.splines import extrap_options as xto
from scipy.ndimage import gaussian_filter

from amset.constants import numeric_types
from amset.electronic_structure.kpoints import (
    get_mesh_from_kpoint_numbers,
    kpoints_to_first_bz,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

from amset.util import array_from_buffer, create_shared_array

logger = logging.getLogger(__name__)


class PeriodicLinearInterpolator(object):
    def __init__(self, nbands, data_shape, interpolators, *args):
        self.nbands = nbands
        self.interpolators = interpolators
        self.data_shape = data_shape

    @classmethod
    def from_data(cls, kpoints, data, gaussian=None):
        grid_kpoints, mesh_dim, sort_idx = cls._grid_kpoints(kpoints)
        nbands, data_shape, interpolators = cls._setup_interpolators(
            data, grid_kpoints, mesh_dim, sort_idx, gaussian
        )
        return cls(nbands, data_shape, interpolators)

    def to_reference(self):
        interpolator_references = self._interpolators_to_reference()
        return self.nbands, self.data_shape, interpolator_references

    def _interpolators_to_reference(self):
        interpolator_references = {}
        for spin, (grid, grid_data) in self.interpolators.items():
            grid_data_buffer, grid_data_shared = create_shared_array(
                grid_data, return_shared_data=True
            )
            self.interpolators[spin] = (grid, grid_data_shared)
            interpolator_references[spin] = (grid, grid_data_buffer)
        return interpolator_references

    @classmethod
    def from_reference(cls, nbands, data_shape, interpolator_references, *args):
        interpolators = cls._interpolators_from_reference(interpolator_references)
        return cls(nbands, data_shape, interpolators, *args)

    @staticmethod
    def _interpolators_from_reference(interpolator_references):
        interpolators = {}
        for spin, (grid, grid_data_buffer) in interpolator_references.items():
            grid_data = array_from_buffer(grid_data_buffer)
            interpolators[spin] = (grid, grid_data)
        return interpolators

    @staticmethod
    def _setup_interpolators(data, grid_kpoints, mesh_dim, sort_idx, gaussian):
        x = grid_kpoints[:, 0, 0, 0]
        y = grid_kpoints[0, :, 0, 1]
        z = grid_kpoints[0, 0, :, 2]

        nbands = {s: c.shape[0] for s, c in data.items()}
        interpolators = {}
        data_shape = None
        for spin, spin_data in data.items():
            data_shape = spin_data.shape[2:]
            spin_nbands = nbands[spin]

            # sort the data then reshape them into the grid. The data
            # can now be indexed as data[iband][ikx][iky][ikz]
            sorted_data = spin_data[:, sort_idx]
            grid_shape = (spin_nbands,) + mesh_dim + data_shape
            grid_data = sorted_data.reshape(grid_shape)

            # wrap the data to account for PBC
            pad_size = ((0, 0), (1, 1), (1, 1), (1, 1)) + ((0, 0),) * len(data_shape)
            grid_data = np.pad(grid_data, pad_size, mode="wrap")

            if gaussian:
                for i in range(len(grid_data)):
                    grid_data[i] = gaussian_filter(
                        grid_data[i], sigma=gaussian, mode="wrap"
                    )

            if spin_nbands == 1:
                # this can cause a bug in RegularGridInterpolator. Have to fake
                # having at least two bands
                spin_nbands = 2
                grid_data = np.tile(grid_data, (2, 1, 1, 1) + (1,) * len(data_shape))

            grid = UCGrid(
                (0, spin_nbands - 1, spin_nbands),
                (x[0], x[-1], len(x)),
                (y[0], y[-1], len(y)),
                (z[0], z[-1], len(z)),
            )
            # flatten remaining axes
            grid_shape = grid_data.shape[:4] + (-1,)
            interpolators[spin] = (grid, grid_data.reshape(grid_shape))

        return nbands, data_shape, interpolators

    @staticmethod
    def _grid_kpoints(kpoints):
        # k-points has to cover the full BZ
        kpoints = kpoints_to_first_bz(kpoints)
        mesh_dim = get_mesh_from_kpoint_numbers(kpoints)
        if np.product(mesh_dim) != len(kpoints):
            raise ValueError("K-points do not cover full Brillouin zone.")

        kpoints = np.around(kpoints, 5)

        # get the indices to sort the k-points on the Z, then Y, then X columns
        sort_idx = np.lexsort((kpoints[:, 2], kpoints[:, 1], kpoints[:, 0]))

        # put the kpoints into a 3D grid so that they can be indexed as
        # kpoints[ikx][iky][ikz] = [kx, ky, kz]
        grid_kpoints = kpoints[sort_idx].reshape(mesh_dim + (3,))

        # Expand the k-point mesh to account for periodic boundary conditions
        grid_kpoints = np.pad(
            grid_kpoints, ((1, 1), (1, 1), (1, 1), (0, 0)), mode="wrap"
        )
        grid_kpoints[0, :, :] -= [1, 0, 0]
        grid_kpoints[:, 0, :] -= [0, 1, 0]
        grid_kpoints[:, :, 0] -= [0, 0, 1]
        grid_kpoints[-1, :, :] += [1, 0, 0]
        grid_kpoints[:, -1, :] += [0, 1, 0]
        grid_kpoints[:, :, -1] += [0, 0, 1]
        return grid_kpoints, mesh_dim, sort_idx

    def interpolate(self, spin, bands, kpoints):
        v = np.concatenate([np.asarray(bands)[:, None], np.asarray(kpoints)], axis=1)

        grid, data = self.interpolators[spin]
        if np.iscomplexobj(data):
            # only allows interpolating floats, so have to separate real and imag
            interp_data = np.empty((len(v),) + self.data_shape, dtype=np.complex)
            interp_data.real = eval_linear(grid, data.real, v, xto.LINEAR).reshape(
                -1, *self.data_shape
            )
            interp_data.imag = eval_linear(grid, data.imag, v, xto.LINEAR).reshape(
                -1, *self.data_shape
            )
        else:
            interp_data = eval_linear(grid, data, v, xto.LINEAR).reshape(
                -1, *self.data_shape
            )

        return interp_data


def group_bands_and_kpoints(band_a, kpoint_a, band_b, kpoint_b):
    kpoint_a = np.asarray(kpoint_a)
    kpoint_b = np.asarray(kpoint_b)

    return_single = False
    if isinstance(band_b, numeric_types):
        # only one band index given
        if len(kpoint_b.shape) > 1:
            # multiple k-point indices given
            band_b = np.array([band_b] * len(kpoint_b))
        else:
            band_b = np.array([band_b])
            kpoint_b = [kpoint_b]
            return_single = True

    else:
        band_b = np.asarray(band_b)

    bands = np.concatenate([[band_a], band_b])
    kpoints = np.concatenate([[kpoint_a], kpoint_b])

    return bands, kpoints, return_single



"""
This module implements methods to calculate electron scattering.
"""

import logging
import multiprocessing
import time
import traceback
from multiprocessing import cpu_count
from queue import Empty
from typing import Any, Dict, List, Optional, Union

import numba
import numpy as np
import quadpy
from pymatgen import Spin
from pymatgen.util.coord import pbc_diff
from scipy.interpolate import griddata

from amset.constants import (
    boltzmann_au,
    defaults,
    ev_to_hartree,
    hbar,
    small_val,
    spin_name,
)
from amset.core.data import AmsetData
from amset.electronic_structure.fd import fd
from amset.electronic_structure.kpoints import kpoints_to_first_bz
from amset.electronic_structure.tetrahedron import (
    TetrahedralBandStructure,
    get_cross_section_values,
    get_projected_intersections,
)
from amset.interpolation.momentum import MRTACalculator
from amset.interpolation.projections import ProjectionOverlapCalculator
from amset.interpolation.wavefunction import WavefunctionOverlapCalculator
from amset.log import log_list, log_time_taken
from amset.scattering.basic import AbstractBasicScattering
from amset.scattering.elastic import (
    AbstractElasticScattering,
    AcousticDeformationPotentialScattering,
)
from amset.scattering.inelastic import AbstractInelasticScattering
from amset.util import (
    create_shared_dict_array,
    dict_array_from_buffer,
    get_progress_bar,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)

_all_scatterers: Union = (
    AbstractElasticScattering.__subclasses__()
    + AbstractInelasticScattering.__subclasses__()
    + AbstractBasicScattering.__subclasses__()
)
_scattering_mechanisms = {m.name: m for m in _all_scatterers}
basic_scatterers = [i.name for i in AbstractBasicScattering.__subclasses__()]

ni = {
    "high": {
        "triangle": quadpy.t2.schemes["xiao_gimbutas_50"](),
        "quad": quadpy.c2.schemes["sommariva_50"](),
    },
    "medium": {
        "triangle": quadpy.t2.schemes["xiao_gimbutas_06"](),
        "quad": quadpy.c2.schemes["sommariva_06"](),
    },
    "low": {
        "triangle": quadpy.t2.schemes["centroid"](),
        "quad": quadpy.c2.schemes["dunavant_00"](),
    },
}

# ni = {
#     "high": {"triangle": quadpy.t2.centroid(), "quad": quadpy.c2.dunavant_00()},
#     "medium": {"triangle": quadpy.t2.centroid(), "quad": quadpy.c2.dunavant_00()},
#     "low": {"triangle": quadpy.t2.centroid(), "quad": quadpy.c2.dunavant_00()},
# }


class ScatteringCalculator(object):
    def __init__(
        self,
        settings: Dict[str, float],
        amset_data: AmsetData,
        cutoff_pad: float,
        scattering_type: Union[str, List[str], float] = "auto",
        nworkers: int = defaults["nworkers"],
        progress_bar: bool = defaults["print_log"],
        cache_wavefunction: bool = defaults["cache_wavefunction"],
    ):
        if amset_data.temperatures is None or amset_data.doping is None:
            raise RuntimeError(
                "AmsetData doesn't contain doping levels or temperatures"
            )

        self.scattering_type = scattering_type
        self.settings = settings
        self.nworkers = nworkers if nworkers != -1 else cpu_count()
        self.scatterers = self.get_scatterers(scattering_type, settings, amset_data)
        self.amset_data = amset_data
        self.progress_bar = progress_bar
        self.cache_wavefunction = cache_wavefunction

        buf = 0.05 * ev_to_hartree
        if self.amset_data.fd_cutoffs:
            self.scattering_energy_cutoffs = (
                self.amset_data.fd_cutoffs[0] - buf,
                self.amset_data.fd_cutoffs[1] + buf,
            )
        else:
            self.scattering_energy_cutoffs = (
                min(self.amset_data.dos.energies) - buf,
                max(self.amset_data.dos.energies) + buf,
            )

        self._coeffs = None
        self._coeffs_mapping = None
        # if only basic scatterers then no need to cache overlaps
        basic_only = len(self.elastic_scatterers) + len(self.inelastic_scatterers) == 0
        if cache_wavefunction and not basic_only:
            self._coeffs = {}
            self._coeffs_mapping = {}
            # precompute the coefficients we will need to for calculating overlaps
            # could do this on the fly but caching will really speed things up.
            # we need to interpolate as the wavefunction coefficients were calculated on
            # a coarse mesh but we calculate the orbital overlap on a fine mesh.
            tbs = self.amset_data.tetrahedral_band_structure
            for spin in amset_data.spins:
                spin_b_idxs = []
                spin_k_idxs = []
                for b_idx, b_energies in enumerate(self.amset_data.energies[spin]):
                    # find all k-points that fall inside Fermi cutoffs
                    k_idxs = np.where(
                        (b_energies > self.scattering_energy_cutoffs[0] - cutoff_pad)
                        & (b_energies < self.scattering_energy_cutoffs[1] + cutoff_pad)
                    )[0]

                    # find k-points connected to the k-points inside Fermi cutoffs
                    k_idxs = tbs.get_connected_kpoints(k_idxs)

                    spin_k_idxs.extend(k_idxs.tolist())
                    spin_b_idxs.extend([b_idx] * len(k_idxs))

                # calculate the coefficients for all bands and k-point simultaneously
                try:
                    self._coeffs[
                        spin
                    ] = self.amset_data.overlap_calculator.get_coefficients(
                        spin, spin_b_idxs, self.amset_data.kpoints[spin_k_idxs]
                    )
                    # because we are only storing the coefficients for the
                    # band/k-points we want, we need a way of mapping from the original
                    # band/k-point indices to the reduced indices. I.e., it allows us to
                    # get the coefficients for band b_idx, and k-point k_idx using:
                    # self._coeffs[spin][self._coeffs_mapping[b_idx, k_idx]]
                    # use a default value of 100000 as this was it will throw an error
                    # if we don't precache the correct values
                    mapping = np.full_like(self.amset_data.energies[spin], 100000)
                    mapping[spin_b_idxs, spin_k_idxs] = np.arange(len(spin_b_idxs))
                    self._coeffs_mapping[spin] = mapping.astype(int)

                except np.core._exceptions.MemoryError:
                    logger.warning(
                        "Memory requirements too large to cache wavefunction "
                        "coefficients. Setting cache_wavefunction to False"
                    )
                    self._coeffs = None
                    self._coeffs_mapping = None
                    break

        self.in_queue = None
        self.out_queue = None
        self.workers = None
        self.initialize_workers()

    def initialize_workers(self):
        logger.info(f"Forking {self.nworkers} processes to calculate scattering")
        t0 = time.perf_counter()

        if isinstance(self.amset_data.overlap_calculator, ProjectionOverlapCalculator):
            overlap_type = "projection"
        else:
            overlap_type = "wavefunction"

        if self._coeffs is None:
            coeffs_buffer = None
            coeffs_mapping_buffer = None
        else:
            coeffs_buffer, self._coeffs = create_shared_dict_array(
                self._coeffs, return_shared_data=True
            )
            coeffs_mapping_buffer, self._coeffs_mapping = create_shared_dict_array(
                self._coeffs_mapping, return_shared_data=True
            )

        amset_data_min = _AmsetDataMin.from_amset_data(self.amset_data)
        amset_data_min_reference = amset_data_min.to_reference()

        # deformation potential is a large tensor that should be put into shared memory
        elastic_scatterers = [
            s.to_reference()
            if isinstance(s, AcousticDeformationPotentialScattering)
            else s
            for s in self.elastic_scatterers
        ]

        ctx = multiprocessing.get_context("spawn")
        self.in_queue = ctx.Queue()
        self.out_queue = ctx.Queue()
        args = (
            self.amset_data.tetrahedral_band_structure.to_reference(),
            overlap_type,
            self.amset_data.overlap_calculator.to_reference(),
            self.amset_data.mrta_calculator.to_reference(),
            elastic_scatterers,
            self.inelastic_scatterers,
            amset_data_min_reference,
            coeffs_buffer,
            coeffs_mapping_buffer,
            self.in_queue,
            self.out_queue,
        )

        self.workers = []
        for _ in range(self.nworkers):
            self.workers.append(ctx.Process(target=scattering_worker, args=args))

        iterable = self.workers
        if self.progress_bar:
            iterable = get_progress_bar(self.workers, desc="workers")

        for w in iterable:
            w.start()

        log_time_taken(t0)
        return self.workers

    def terminate_workers(self):
        # The "None"s at the end of the queue signal the workers that there are
        # no more jobs left and they must therefore exit.
        for i in range(self.nworkers):
            self.in_queue.put(None)

        for w in self.workers:
            w.join(0)
            w.terminate()

        self.workers = None

    @property
    def basic_scatterers(self):
        return [s for s in self.scatterers if isinstance(s, AbstractBasicScattering)]

    @property
    def inelastic_scatterers(self):
        return [
            s for s in self.scatterers if isinstance(s, AbstractInelasticScattering)
        ]

    @property
    def elastic_scatterers(self):
        return [s for s in self.scatterers if isinstance(s, AbstractElasticScattering)]

    @property
    def scatterer_labels(self):
        basic_names = [s.name for s in self.basic_scatterers]
        elastic_names = [s.name for s in self.elastic_scatterers]
        inelastic_names = [s.name for s in self.inelastic_scatterers]

        return basic_names + elastic_names + inelastic_names

    @staticmethod
    def get_scatterers(
        scattering_type: Union[str, List[str], float],
        settings: Dict[str, Any],
        amset_data: AmsetData,
    ) -> List[Union[AbstractElasticScattering, AbstractInelasticScattering]]:
        if scattering_type == "auto":
            # dynamically determine the available scattering mechanism subclasses
            logger.info(
                "Examining material properties to determine possible scattering "
                "mechanisms"
            )

            scattering_type = []
            for name, mechanism in _scattering_mechanisms.items():
                req_prop = mechanism.required_properties
                if all([settings.get(x, None) is not None for x in req_prop]):
                    scattering_type.append(name)

            if not scattering_type:
                raise ValueError(
                    "No scattering mechanisms possible with material properties"
                )

        else:
            for name in scattering_type:
                missing_properties = [
                    p
                    for p in _scattering_mechanisms[name].required_properties
                    if settings.get(p, None) is None
                ]

                if missing_properties:
                    str_missing = ", ".join(missing_properties)
                    raise ValueError(
                        "{} scattering mechanism specified but the following "
                        "material properties are missing: {}".format(name, str_missing)
                    )

        str_scats = ", ".join(scattering_type)
        logger.info("Scattering mechanisms to be calculated: {}".format(str_scats))

        return [
            _scattering_mechanisms[name].from_amset_data(settings, amset_data)
            for name in scattering_type
        ]

    def calculate_scattering_rates(self):
        spins = self.amset_data.spins
        kpoints = self.amset_data.kpoints
        energies = self.amset_data.energies
        fermi_shape = self.amset_data.fermi_levels.shape
        scattering_shape = (len(self.scatterer_labels),) + fermi_shape
        rate_shape = {s: scattering_shape + energies[s].shape for s in spins}

        # rates has shape (spin, nscatterers, ndoping, ntemp, nbands, nkpoints)
        rates = {s: np.zeros(rate_shape[s]) for s in spins}
        masks = {s: np.full(rate_shape[s], True) for s in spins}

        logger.info("Scattering information:")
        log_list(["# ir k-points: {}".format(len(self.amset_data.ir_kpoints_idx))])

        for spin in spins:
            for b_idx in range(len(self.amset_data.energies[spin])):
                str_b = "Calculating rates for {} band {}"
                logger.info(str_b.format(spin_name[spin], b_idx + 1))

                t0 = time.perf_counter()
                (
                    rates[spin][..., b_idx, :],
                    masks[spin][..., b_idx, :],
                ) = self.calculate_band_rates(spin, b_idx)

                info = [
                    "max rate: {:.4g}".format(rates[spin][..., b_idx, :].max()),
                    "min rate: {:.4g}".format(rates[spin][..., b_idx, :].min()),
                ]
                log_list(info, level=logging.DEBUG)
                log_list(["time: {:.4f} s".format(time.perf_counter() - t0)])

            # fill in k-points outside Fermi-Dirac cutoffs with a default value
            rates[spin][masks[spin]] = 1e14

        # if the k-point density is low, some k-points may not have other k-points
        # within the energy tolerance leading to zero rates
        rates = _interpolate_zero_rates(
            rates, kpoints, masks, progress_bar=self.progress_bar
        )

        # enforce symmetry of interpolated points
        ir_idx = self.amset_data.ir_kpoints_idx
        ir_to_full = self.amset_data.ir_to_full_kpoint_mapping
        rates = {s: sr[..., ir_idx][..., ir_to_full] for s, sr in rates.items()}

        self.terminate_workers()
        return rates

    def calculate_band_rates(self, spin: Spin, b_idx: int):
        if self.workers is None:
            self.initialize_workers()

        vol = self.amset_data.structure.lattice.reciprocal_lattice.volume
        conversion = vol / (4 * np.pi ** 2)
        kpoints_idx = self.amset_data.ir_kpoints_idx
        nkpoints = len(kpoints_idx)

        band_energies = self.amset_data.energies[spin][b_idx, kpoints_idx]
        mask = band_energies < self.scattering_energy_cutoffs[0]
        mask |= band_energies > self.scattering_energy_cutoffs[1]
        fill_mask = mask[self.amset_data.ir_to_full_kpoint_mapping]

        n = np.sum(~fill_mask)
        logger.info("  ├── # k-points within Fermi–Dirac cut-offs: {}".format(n))

        k_idx_in_cutoff = kpoints_idx[~mask]
        ir_idx_in_cutoff = np.arange(nkpoints)[~mask]
        iterable = list(zip(k_idx_in_cutoff, ir_idx_in_cutoff))

        to_stack = []
        if len(self.basic_scatterers) > 0:
            basic_rates = np.array(
                [m.rates[spin][:, :, b_idx, kpoints_idx] for m in self.basic_scatterers]
            )
            to_stack.append(basic_rates)

        if len(self.elastic_scatterers) > 0:
            elastic_prefactors = conversion * np.array(
                [m.prefactor(spin, b_idx) for m in self.elastic_scatterers]
            )
            elastic_rates = np.zeros(elastic_prefactors.shape + (nkpoints,))

            if len(k_idx_in_cutoff) > 0:
                if self.progress_bar:
                    pbar = get_progress_bar(total=len(iterable), desc="elastic")
                else:
                    pbar = None

                for k_idx, ir_idx in iterable:
                    self.in_queue.put((spin, b_idx, k_idx, False, ir_idx))

                for _ in range(len(iterable)):
                    ir_idx, elastic_rates[..., ir_idx] = self._get_rate_from_queue()
                    if pbar:
                        pbar.update()

                if pbar:
                    pbar.close()

            elastic_rates *= elastic_prefactors[..., None]
            to_stack.append(elastic_rates)

        if len(self.inelastic_scatterers) > 0:
            inelastic_prefactors = conversion * np.array(
                [m.prefactor(spin, b_idx) for m in self.inelastic_scatterers]
            )
            inelastic_rates = np.zeros(inelastic_prefactors.shape + (nkpoints,))
            f_pop = self.settings["pop_frequency"]
            energy_diff = f_pop * 1e12 * 2 * np.pi * hbar * ev_to_hartree

            if len(k_idx_in_cutoff) > 0:
                if self.progress_bar:
                    pbar = get_progress_bar(total=len(iterable) * 2, desc="inelastic")
                else:
                    pbar = None

                inelastic_rates[:, :, :, ir_idx_in_cutoff] = 0
                for k_idx, ir_idx in iterable:
                    for ediff in [energy_diff, -energy_diff]:
                        self.in_queue.put((spin, b_idx, k_idx, ediff, ir_idx))

                for i in range(len(iterable) * 2):
                    ir_idx, rate = self._get_rate_from_queue()
                    inelastic_rates[..., ir_idx] += rate
                    if pbar:
                        pbar.update()

                if pbar:
                    pbar.close()

            inelastic_rates *= inelastic_prefactors[..., None]
            to_stack.append(inelastic_rates)

        all_band_rates = np.vstack(to_stack)

        return all_band_rates[..., self.amset_data.ir_to_full_kpoint_mapping], fill_mask

    def _get_rate_from_queue(self):
        # handle exception gracefully to avoid hanging processes
        try:
            result = self.out_queue.get(timeout=10)
        except Empty:
            # didn't receive anything for 10 seconds; this could be OK or it could
            # the processes have been killed
            if not self._workers_alive():
                self.terminate_workers()
                raise MemoryError(
                    "Some subprocessess were killed unexpectedly. Could be OOM "
                    "Killer?\nTry reducing nworkers."
                )
            else:
                return self._get_rate_from_queue()

        if isinstance(result[0], Exception):
            logger.error(
                "Scattering process ended with error:\n{}\nexiting".format(
                    str(result[1])
                )
            )
            self.terminate_workers()
            raise result[0]
        return result

    def _workers_alive(self):
        return all([worker.is_alive() for worker in self.workers])


def scattering_worker(
    tbs_reference,
    overlap_type,
    overlap_calculator_reference,
    mrta_calculator_reference,
    elastic_scatterers,
    inelastic_scatterers,
    amset_data_min_reference,
    coeffs_buffer,
    coeffs_mapping_buffer,
    in_queue,
    out_queue,
):
    try:
        tbs = TetrahedralBandStructure.from_reference(*tbs_reference)
        mrta_calculator = MRTACalculator.from_reference(*mrta_calculator_reference)
        amset_data_min = _AmsetDataMin.from_reference(*amset_data_min_reference)

        if coeffs_buffer is None:
            coeffs = None
            coeffs_mapping = None
        else:
            coeffs = dict_array_from_buffer(coeffs_buffer)
            coeffs_mapping = dict_array_from_buffer(coeffs_mapping_buffer)

        overlap_calculator = None
        if overlap_type == "wavefunction":
            overlap_calculator = WavefunctionOverlapCalculator.from_reference(
                *overlap_calculator_reference
            )
        elif overlap_type == "projection":
            overlap_calculator = ProjectionOverlapCalculator.from_reference(
                *overlap_calculator_reference
            )
        else:
            raise ValueError("Unrecognised overlap type: {}".format(overlap_type))

        elastic_scatterers = [
            AcousticDeformationPotentialScattering.from_reference(*s)
            if isinstance(s, tuple)
            else s
            for s in elastic_scatterers
        ]

        with np.errstate(all="ignore"):
            while True:
                job = in_queue.get()

                if job is None:
                    break

                spin, b_idx, k_idx, energy_diff, ir_k_idx = job
                rate = calculate_rate(
                    tbs,
                    overlap_calculator,
                    mrta_calculator,
                    elastic_scatterers,
                    inelastic_scatterers,
                    amset_data_min,
                    coeffs,
                    coeffs_mapping,
                    spin,
                    b_idx,
                    k_idx,
                    energy_diff=energy_diff,
                )
                out_queue.put((ir_k_idx, rate))

    except BaseException as e:
        error_msg = traceback.format_exc()
        out_queue.put((e, error_msg))


class _AmsetDataMin:
    def __init__(self, structure, kpoint_mesh, velocities, fermi_levels, temperatures):
        self.structure = structure
        self.kpoint_mesh = kpoint_mesh
        self.velocities = velocities
        self.fermi_levels = fermi_levels
        self.temperatures = temperatures

    def to_reference(self):
        velocities_buffer, self.velocities = create_shared_dict_array(
            self.velocities, return_shared_data=True
        )
        return (
            self.structure,
            self.kpoint_mesh,
            velocities_buffer,
            self.fermi_levels,
            self.temperatures,
        )

    @classmethod
    def from_reference(
        cls, structure, kpoint_mesh, velocities_buffer, fermi_levels, temperatures
    ):
        return cls(
            structure,
            kpoint_mesh,
            dict_array_from_buffer(velocities_buffer),
            fermi_levels,
            temperatures,
        )

    @classmethod
    def from_amset_data(cls, amset_data):
        return cls(
            amset_data.structure,
            amset_data.kpoint_mesh,
            amset_data.velocities,
            amset_data.fermi_levels,
            amset_data.temperatures,
        )


def calculate_rate(
    tbs: TetrahedralBandStructure,
    overlap_calculator,
    mrta_calculator,
    elastic_scatterers,
    inelastic_scatterers,
    amset_data_min: _AmsetDataMin,
    coeffs,
    coeffs_mapping,
    spin,
    b_idx,
    k_idx,
    energy_diff=None,
):
    rlat = amset_data_min.structure.lattice.reciprocal_lattice.matrix
    velocity = amset_data_min.velocities[spin][b_idx, k_idx]
    energy = tbs.energies[spin][b_idx, k_idx]

    if energy_diff:
        energy += energy_diff

    (
        tet_dos,
        tet_mask,
        cs_weights,
        tet_contributions,
    ) = tbs.get_tetrahedra_density_of_states(
        spin,
        energy,
        return_contributions=True,
        symmetry_reduce=False,
        # band_idx=b_idx,  # turn this on to disable interband scattering
    )

    if len(tet_dos) == 0:
        return 0

    # next, get k-point indices and band_indices
    property_mask, band_kpoint_mask, band_mask, kpoint_mask = tbs.get_masks(
        spin, tet_mask
    )
    k = tbs.kpoints[k_idx]
    k_primes = tbs.kpoints[kpoint_mask]

    if coeffs is not None:
        # use cached coefficients to calculate the overlap on the fine mesh
        # tetrahedron vertices
        spin_coeffs = coeffs[spin]
        spin_coeffs_mapping = coeffs_mapping[spin]
        if len(spin_coeffs.shape) == 3:
            # ncl
            overlap = _get_overlap_ncl(
                spin_coeffs, spin_coeffs_mapping, b_idx, k_idx, band_mask, kpoint_mask
            )
        else:
            overlap = _get_overlap(
                spin_coeffs, spin_coeffs_mapping, b_idx, k_idx, band_mask, kpoint_mask
            )
    else:
        overlap = overlap_calculator.get_overlap(spin, b_idx, k, band_mask, k_primes)

    # put overlap back in array with shape (nbands, nkpoints)
    all_overlap = np.zeros(tbs.energies[spin].shape)
    all_overlap[band_kpoint_mask] = overlap

    # now select the properties at the tetrahedron vertices
    vert_overlap = all_overlap[property_mask]

    # get interpolated overlap at centre of tetrahedra cross sections
    tet_overlap = get_cross_section_values(vert_overlap, *tet_contributions)
    tetrahedra = tbs.tetrahedra[spin][tet_mask]

    # have to deal with the case where the tetrahedron cross section crosses the
    # zone boundary. This is a slight inaccuracy but we just treat the
    # cross section as if it is on one side of the boundary
    tet_kpoints = tbs.kpoints[tetrahedra]
    base_kpoints = tet_kpoints[:, 0][:, None, :]
    k_diff = pbc_diff(tet_kpoints, base_kpoints) + pbc_diff(base_kpoints, k)

    # project the tetrahedron cross sections onto 2D surfaces in either a triangle
    # or quadrilateral
    k_diff = np.dot(k_diff, rlat)
    intersections = get_cross_section_values(k_diff, *tet_contributions, average=False)
    projected_intersections, basis = get_projected_intersections(intersections)

    k_spacing = np.linalg.norm(np.dot(rlat, 1 / amset_data_min.kpoint_mesh))
    qpoints, weights, mapping = get_fine_mesh_qpoints(
        projected_intersections,
        basis,
        *tet_contributions[0:3],
        high_tol=k_spacing * 0.5,
        med_tol=k_spacing * 2,
        cross_section_weights=cs_weights,
    )
    qpoint_norm_sq = np.sum(qpoints ** 2, axis=-1)

    k_primes = np.dot(qpoints, np.linalg.inv(rlat)) + k
    k_primes = kpoints_to_first_bz(k_primes)

    # unit q in reciprocal cartesian coordinates
    unit_q = qpoints / np.sqrt(qpoint_norm_sq)[:, None]
    if energy_diff:
        e_fd = _get_fd(energy, amset_data_min.fermi_levels, amset_data_min.temperatures)
        emission = energy_diff <= 0
        rates = [
            s.factor(unit_q, qpoint_norm_sq, emission, e_fd)
            for s in inelastic_scatterers
        ]
        mrta_factor = 1
    else:
        mrta_factor = mrta_calculator.get_mrta_factor(
            spin, b_idx, k, tet_mask[0][mapping], k_primes
        )
        rates = [
            s.factor(unit_q, qpoint_norm_sq, spin, b_idx, k, velocity)
            for s in elastic_scatterers
        ]

    rates = np.array(rates)
    rates /= amset_data_min.structure.lattice.reciprocal_lattice.volume
    rates *= tet_overlap[mapping] * weights * mrta_factor

    # this is too expensive vs tetrahedron integration and doesn't add much more
    # accuracy; could offer this as an option
    # overlap = self.amset_data.overlap_calculator.get_overlap(
    #     spin, b_idx, k, tet_mask[0][mapping], k_primes
    # )
    # rates *= overlap * weights * mrta_factor

    # sometimes the projected intersections can be nan when the density of states
    # contribution is infinitesimally small; this catches those errors
    rates[np.isnan(rates)] = 0

    return np.sum(rates, axis=-1)


@numba.njit
def _get_overlap(
    spin_coeffs, spin_coeffs_mapping, b_idx, k_idx, band_mask, kpoint_mask
):
    res = np.zeros(band_mask.shape[0])
    initial = np.conj(spin_coeffs[spin_coeffs_mapping[b_idx, k_idx]])

    for i in range(band_mask.shape[0]):
        final = spin_coeffs[spin_coeffs_mapping[band_mask[i], kpoint_mask[i]]]
        res[i] = np.abs(np.dot(final, initial)) ** 2

    return res


@numba.njit
def _get_overlap_ncl(
    spin_coeffs, spin_coeffs_mapping, b_idx, k_idx, band_mask, kpoint_mask
):
    res = np.zeros(band_mask.shape[0])
    initial = np.conj(spin_coeffs[spin_coeffs_mapping[b_idx, k_idx]])

    for i in range(band_mask.shape[0]):
        final = spin_coeffs[spin_coeffs_mapping[band_mask[i], kpoint_mask[i]]]
        sum_ = 0j
        for j in range(final.shape[0]):
            sum_ += initial[j, 0] * final[j, 0] + initial[j, 1] + final[j, 1]
        res[i] = abs(sum_) ** 2

    return res


def _interpolate_zero_rates(
    rates, kpoints, masks: Optional = None, progress_bar: bool = defaults["print_log"]
):
    # loop over all scattering types, doping, temps, and bands and interpolate
    # zero scattering rates based on the nearest k-point
    logger.info("Interpolating missing scattering rates")
    n_rates = sum([np.product(rates[spin].shape[:-1]) for spin in rates])
    if progress_bar:
        pbar = get_progress_bar(total=n_rates, desc="progress")
    else:
        pbar = None

    t0 = time.perf_counter()
    k_idx = np.arange(len(kpoints))
    for spin in rates:
        for s, d, t, b in np.ndindex(rates[spin].shape[:-1]):

            if masks is not None:
                mask = np.invert(masks[spin][s, d, t, b])
            else:
                mask = [True] * len(rates[spin][s, d, t, b])

            non_zero_rates = rates[spin][s, d, t, b, mask] > 1e7
            # non_zero_rates = rates[spin][s, d, t, b, mask] != 0
            zero_rate_idx = k_idx[mask][~non_zero_rates]
            non_zero_rate_idx = k_idx[mask][non_zero_rates]

            if not np.any(non_zero_rates):
                # all scattering rates are zero so cannot interpolate
                # generally this means the scattering prefactor is zero. E.g.
                # for POP when studying non polar materials
                rates[spin][s, d, t, b, mask] += small_val

            elif np.sum(non_zero_rates) != np.sum(mask):
                # seems to work best when all the kpoints are +ve therefore add 0.5
                # Todo: Use cartesian coordinates?
                # interpolate log rates to avoid the bias towards large rates
                rates[spin][s, d, t, b, zero_rate_idx] = np.exp(
                    griddata(
                        points=kpoints[non_zero_rate_idx] + 0.5,
                        values=np.log(rates[spin][s, d, t, b, non_zero_rate_idx]),
                        xi=kpoints[zero_rate_idx] + 0.5,
                        method="nearest",
                    )
                )
                # rates[spin][s, d, t, b, zero_rate_idx] = 1e15

            if pbar is not None:
                pbar.update()

    if pbar is not None:
        pbar.close()
    log_time_taken(t0)

    return rates


def get_fine_mesh_qpoints(
    intersections,
    basis,
    cond_a_mask,
    cond_b_mask,
    cond_c_mask,
    high_tol=0.1,
    med_tol=0.2,
    cross_section_weights=None,
):
    if cross_section_weights is None:
        cross_section_weights = np.ones(len(intersections))

    # minimum norm in each intersection
    all_norms = np.linalg.norm(intersections, axis=-1)

    # intersections now has shape nvert, ntet, 2 (i.e., x, y coords)
    intersection_idxs = np.arange(len(intersections))
    z_coords = intersections[:, 0, 2]
    intersections = intersections[:, :, :2].transpose(1, 0, 2)
    triangle_mask = cond_a_mask | cond_c_mask

    # have to do these separately as the triangle intersections always have [0, 0, 0]
    # as the last coordinate
    norms = np.ones(len(all_norms))
    norms[triangle_mask] = np.min(all_norms[:, :3][triangle_mask], axis=-1)
    norms[cond_b_mask] = np.min(all_norms[cond_b_mask], axis=-1)

    qpoints = []
    qweights = []
    mapping = []  # maps a qpoint to an intersection index

    def _get_tri_mesh(prec, min_norm, max_norm):
        scheme = ni[prec]["triangle"]
        mask = (min_norm <= norms) & (norms < max_norm) & triangle_mask
        if not np.any(mask):
            return

        simplex = intersections[:3, mask]
        vol = quadpy.tn.get_vol(simplex)
        xy_coords = quadpy.tn.transform(scheme.points, simplex.T)
        weights = (
            scheme.weights[None] * vol[:, None] * cross_section_weights[mask][:, None]
        )

        qpoints.append(get_q(xy_coords, z_coords[mask]))
        qweights.append(weights.reshape(-1))
        mapping.append(np.repeat(intersection_idxs[mask], len(scheme.weights)))

    def _get_quad_mesh(prec, min_norm, max_norm):
        scheme = ni[prec]["quad"]
        mask = (min_norm <= norms) & (norms < max_norm) & cond_b_mask
        if not np.any(mask):
            return

        cube = intersections.reshape((2, 2, -1, 2))[:, :, mask]
        # 4 is taken from quadpy CnScheme.integrate
        # ref_vol = 2 ** numpy.prod(len(ncube.shape) - 1) which for quadrilaterals = 4
        vol = 4 * np.abs(quadpy.cn._helpers.get_detJ(scheme.points, cube))
        xy_coords = quadpy.cn.transform(scheme.points, cube).T
        weights = scheme.weights[None] * vol * cross_section_weights[mask][:, None]

        qpoints.append(get_q(xy_coords, z_coords[mask]))
        qweights.append(weights.reshape(-1))
        mapping.append(np.repeat(intersection_idxs[mask], len(scheme.weights)))

    _get_tri_mesh("high", 0, high_tol)
    _get_tri_mesh("medium", high_tol, med_tol)
    _get_tri_mesh("low", med_tol, np.Inf)
    _get_quad_mesh("high", 0, high_tol)
    _get_quad_mesh("medium", high_tol, med_tol)
    _get_quad_mesh("low", med_tol, np.Inf)

    qpoints = np.concatenate(qpoints)
    qweights = np.concatenate(qweights)
    mapping = np.concatenate(mapping)

    return get_kpoints_in_original_basis(qpoints, basis[mapping]), qweights, mapping


def get_kpoints_in_original_basis(q, basis):
    # transform k back to original lattice basis in cartesian coords
    return np.einsum("ikj,ij->ik", basis, q)


def get_q(x, z_coords):
    z = np.repeat(z_coords[:, None], len(x[0][0]), axis=-1)
    return np.stack([x[0], x[1], z], axis=-1).reshape(-1, 3)


def _get_fd(energy, fermi_levels, temperatures):
    f = np.zeros(fermi_levels.shape)

    for n, t in np.ndindex(fermi_levels.shape):
        f[n, t] = fd(energy, fermi_levels[n, t], temperatures[t] * boltzmann_au)
    return f



import logging
from typing import List, Tuple, Union

import numpy as np
from pymatgen import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from spglib import spglib

from amset.constants import defaults, ktol

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


def kpoints_to_first_bz(
    kpoints: np.ndarray, tol=ktol, negative_zone_boundary: bool = True
) -> np.ndarray:
    """Translate fractional k-points to the first Brillouin zone.

    I.e. all k-points will have fractional coordinates:
        -0.5 <= fractional coordinates < 0.5

    Args:
        kpoints: The k-points in fractional coordinates.
        tol: Fractional tolerance for evaluating zone boundary points.
        negative_zone_boundary: Whether to use -0.5 (spglib convention) or
            0.5 (VASP convention) for zone boundary points.

    Returns:
        The translated k-points.
    """
    kp = kpoints - np.round(kpoints)

    # account for small rounding errors for 0.5
    round_dp = int(np.log10(1 / tol))
    krounded = np.round(kp, round_dp)

    if negative_zone_boundary:
        kp[krounded == 0.5] = -0.5
    else:
        kp[krounded == -0.5] = 0.5
    return kp


def sort_boltztrap_to_spglib(kpoints):
    sort_idx = np.lexsort(
        (
            kpoints[:, 2],
            kpoints[:, 2] < 0,
            kpoints[:, 1],
            kpoints[:, 1] < 0,
            kpoints[:, 0],
            kpoints[:, 0] < 0,
        )
    )
    boltztrap_kpoints = kpoints[sort_idx]

    sort_idx = np.lexsort(
        (
            boltztrap_kpoints[:, 0],
            boltztrap_kpoints[:, 0] < 0,
            boltztrap_kpoints[:, 1],
            boltztrap_kpoints[:, 1] < 0,
            boltztrap_kpoints[:, 2],
            boltztrap_kpoints[:, 2] < 0,
        )
    )
    return sort_idx


def get_kpoints_tetrahedral(
    kpoint_mesh: Union[float, List[int]],
    structure: Structure,
    symprec: float = defaults["symprec"],
    time_reversal_symmetry: bool = True,
) -> Tuple[np.ndarray, ...]:
    """Gets the symmetry inequivalent k-points from a k-point mesh.

    Follows the same process as SpacegroupAnalyzer.get_ir_reciprocal_mesh
    but is faster and allows returning of the full k-point mesh and mapping.

    Args:
        kpoint_mesh: The k-point mesh as a 1x3 array. E.g.,``[6, 6, 6]``.
            Alternatively, if a single value is provided this will be
            treated as a k-point spacing cut-off and the k-points will be generated
            automatically.  Cutoff is length in Angstroms and corresponds to
            non-overlapping radius in a hypothetical supercell (Moreno-Soler length
            cutoff).
        structure: A structure.
        symprec: Symmetry tolerance used when determining the symmetry
            inequivalent k-points on which to interpolate.
        time_reversal_symmetry: Whether the system has time reversal symmetry.

    Returns:
        The irreducible k-points and their weights as tuple, formatted as::

            (ir_kpoints, weights)

        If return_full_kpoints, the data will be returned as::

            (ir_kpoints, weights, kpoints, ir_kpoints_idx, ir_to_full_idx)

        Where ``ir_kpoints_idx`` is the index of the unique irreducible k-points
        in ``kpoints``. ``ir_to_full_idx`` is a list of indices that can be
        used to construct the full Brillouin zone from the ir_mesh. Note the
        ir -> full conversion will only work with calculated scalar properties
        such as energy (not vector properties such as velocity).
    """
    from amset.electronic_structure.tetrahedron import get_tetrahedra

    if isinstance(kpoint_mesh, (int, float)):
        kpoint_mesh = get_kpoint_mesh(structure, kpoint_mesh)

    atoms = AseAtomsAdaptor().get_atoms(structure)

    if not symprec:
        symprec = 1e-8

    grid_mapping, grid_address = spglib.get_ir_reciprocal_mesh(
        kpoint_mesh, atoms, symprec=symprec, is_time_reversal=time_reversal_symmetry
    )
    full_kpoints = grid_address / kpoint_mesh

    tetra, ir_tetrahedra_idx, ir_tetrahedra_to_full_idx, tet_weights = get_tetrahedra(
        structure.lattice.reciprocal_lattice.matrix,
        grid_address,
        kpoint_mesh,
        grid_mapping,
    )

    ir_kpoints_idx, ir_to_full_idx, weights = np.unique(
        grid_mapping, return_inverse=True, return_counts=True
    )
    ir_kpoints = full_kpoints[ir_kpoints_idx]

    return (
        ir_kpoints,
        weights,
        full_kpoints,
        ir_kpoints_idx,
        ir_to_full_idx,
        tetra,
        ir_tetrahedra_idx,
        ir_tetrahedra_to_full_idx,
        tet_weights,
    )


def get_kpoint_mesh(structure: Structure, cutoff_length: float, force_odd: bool = True):
    """Calculate reciprocal-space sampling with real-space cut-off."""
    reciprocal_lattice = structure.lattice.reciprocal_lattice_crystallographic

    # Get reciprocal cell vector magnitudes
    abc_recip = np.array(reciprocal_lattice.abc)

    mesh = np.ceil(abc_recip * 2 * cutoff_length).astype(int)

    if force_odd:
        mesh += (mesh + 1) % 2

    return mesh


def get_mesh_from_kpoint_diff(kpoints):
    kpoints = np.array(kpoints)

    # whether the k-point mesh is shifted or Gamma centered mesh
    is_shifted = np.min(np.linalg.norm(kpoints, axis=1)) > 1e-6

    unique_a = np.unique(kpoints[:, 0])
    unique_b = np.unique(kpoints[:, 1])
    unique_c = np.unique(kpoints[:, 2])

    if len(unique_a) == 1:
        na = 1
    else:
        na = 1 / np.min(np.diff(unique_a))

    if len(unique_b) == 1:
        nb = 1
    else:
        nb = 1 / np.min(np.diff(unique_b))

    if len(unique_c) == 1:
        nc = 1
    else:
        nc = 1 / np.min(np.diff(unique_c))

    # due to limited precission of the input k-points, the mesh is returned as a float
    return np.array([na, nb, nc]), is_shifted


def get_mesh_from_kpoint_numbers(kpoints, tol=ktol):
    round_dp = int(np.log10(1 / tol))

    kpoints = kpoints_to_first_bz(kpoints)
    round_kpoints = np.round(kpoints, round_dp)

    nx = len(np.unique(round_kpoints[:, 0]))
    ny = len(np.unique(round_kpoints[:, 1]))
    nz = len(np.unique(round_kpoints[:, 2]))

    return nx, ny, nz


def get_kpoint_indices(kpoints, mesh, is_shifted=False):
    mesh = np.array(mesh)
    shift = np.array([1, 1, 1]) if is_shifted else np.array([0, 0, 0])
    min_kpoint = -np.floor(mesh / 2).round().astype(int)
    addresses = ((kpoints + shift / (mesh * 2)) * mesh).round().astype(int)
    shifted = addresses - min_kpoint
    nyz = mesh[1] * mesh[2]
    nz = mesh[2]
    indices = shifted[:, 0] * nyz + shifted[:, 1] * nz + shifted[:, 2]
    return indices.round().astype(int)


def get_kpoints_from_bandstructure(bandstructure, cartesian=False):
    if cartesian:
        return np.array([k.coords for k in bandstructure.kpoints])
    else:
        return np.array([k.frac_coords for k in bandstructure.kpoints])



import numpy as np

from amset.constants import boltzmann_au
from amset.electronic_structure.fd import fd

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


def calculate_inverse_screening_length_sq(amset_data, dielectric):
    inverse_screening_length_sq = np.zeros(amset_data.fermi_levels.shape)

    tdos = amset_data.dos.tdos
    energies = amset_data.dos.energies
    fermi_levels = amset_data.fermi_levels
    vol = amset_data.structure.volume

    for n, t in np.ndindex(inverse_screening_length_sq.shape):
        ef = fermi_levels[n, t]
        temp = amset_data.temperatures[t]
        f = fd(energies, ef, temp * boltzmann_au)
        integral = np.trapz(tdos * f * (1 - f), x=energies)
        inverse_screening_length_sq[n, t] = (
            integral * 4 * np.pi / (dielectric * boltzmann_au * temp * vol)
        )

    return inverse_screening_length_sq



import sys
from typing import Optional, Union

import click
from click import argument, option

from amset.tools.common import echo_ibands, path_type, zero_weighted_type
from amset.util import parse_ibands

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

_symprec = 0.01  # redefine symprec to avoid loading constants from file
_metal_str = {True: "metallic", False: "semiconducting"}
_kpt_str = "[{k[0]:6.2f} {k[1]:6.2f} {k[2]:6.2f}  ]"
_tensor_str = """    [[{:6.2f} {:6.2f} {:6.2f}]
     [{:6.2f} {:6.2f} {:6.2f}]
     [{:6.2f} {:6.2f} {:6.2f}]]"""


def _parse_symprec(var: Optional[Union[str, float, int]]):
    if var is None:
        return _symprec
    if isinstance(var, float):
        return var
    if "N" in var:
        return None
    return float(var)


@click.group()
def deform():
    """
    Tools for calculating acoustic deformation potentials.
    """
    pass


@deform.command()
@option("-f", "--filename", default="POSCAR", help="path to input structure file")
@option(
    "-d", "--distance", type=float, default=0.005, help="fractional magnitude of strain"
)
@option(
    "-s",
    "--symprec",
    help="symmetry precision for reducing deformations (use 'N' for no symmetry)",
)
@option("--directory", type=path_type, help="file output directory")
def create(**kwargs):
    """
    Generate deformed structures for calculating deformation potentials.
    """
    from pymatgen import Structure
    from pymatgen.core.tensors import symmetry_reduce
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
    from pymatgen.util.string import unicodeify_spacegroup

    from amset.deformation.common import get_formatted_tensors
    from amset.deformation.generation import get_deformations, get_deformed_structures
    from amset.deformation.io import write_deformed_poscars

    symprec = _parse_symprec(kwargs["symprec"])

    structure = Structure.from_file(kwargs["filename"])

    click.echo("Generating deformations:")
    click.echo("  - Strain distance: {:g}".format(kwargs["distance"]))

    deformations = get_deformations(kwargs["distance"])
    click.echo("  - # Total deformations: {}".format(len(deformations)))

    if symprec is not None:
        sga = SpacegroupAnalyzer(structure, symprec=symprec)
        spg_symbol = unicodeify_spacegroup(sga.get_space_group_symbol())
        spg_number = sga.get_space_group_number()
        click.echo("  - Spacegroup: {} ({})".format(spg_symbol, spg_number))

        deformations = list(symmetry_reduce(deformations, structure, symprec=symprec))
        click.echo("  - # Inequivalent deformations: {}".format(len(deformations)))

    click.echo("\nDeformations:")
    click.echo("  - " + "\n  - ".join(get_formatted_tensors(deformations)))

    deformed_structures = get_deformed_structures(structure, deformations)

    write_deformed_poscars(deformed_structures, directory=kwargs["directory"])
    click.echo("\nDeformed structures have been created")


@deform.command()
@argument("bulk-folder", type=path_type)
@argument("deformation-folders", nargs=-1, type=path_type)
@option(
    "-s",
    "--symprec",
    help="symmetry precision for reducing deformations (use 'N' for no symmetry)",
)
@option("-e", "--energy-cutoff", type=float, help="energy cutoff for finding bands")
@option(
    "-b",
    "--bands",
    type=str,
    help="bands to calculate the deformation for (overrides energy-cutoff)",
)
@option(
    "-z",
    "--zero-weighted-kpoints",
    help="how to process zero-weighted k-points",
    type=zero_weighted_type,
)
@option("-o", "--output", default="deformation.h5", help="output file path")
def read(bulk_folder, deformation_folders, **kwargs):
    """
    Read deformation calculations and extract deformation potentials.
    """
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
    from pymatgen.util.string import unicodeify_spacegroup

    from amset.constants import defaults
    from amset.deformation.common import get_formatted_tensors
    from amset.deformation.io import parse_calculation, write_deformation_potentials
    from amset.deformation.potentials import (
        calculate_deformation_potentials,
        extract_bands,
        get_strain_mapping,
        get_symmetrized_strain_mapping,
        strain_coverage_ok,
    )
    from amset.electronic_structure.common import get_ibands
    from amset.electronic_structure.kpoints import get_kpoints_from_bandstructure
    from amset.electronic_structure.symmetry import expand_bandstructure

    energy_cutoff = kwargs.pop("energy_cutoff")
    if not energy_cutoff:
        energy_cutoff = defaults["energy_cutoff"]

    zwk_mode = kwargs.pop("zero_weighted_kpoints")
    if not zwk_mode:
        zwk_mode = defaults["zero_weighted_kpoints"]

    symprec = _parse_symprec(kwargs["symprec"])
    click.echo("Reading bulk (undeformed) calculation")
    bulk_calculation = parse_calculation(bulk_folder, zero_weighted_kpoints=zwk_mode)
    bulk_structure = bulk_calculation["bandstructure"].structure

    deformation_calculations = []
    for deformation_folder in deformation_folders:
        click.echo("Reading deformation calculation in {}".format(deformation_folder))
        deformation_calculation = parse_calculation(
            deformation_folder, zero_weighted_kpoints=zwk_mode
        )
        if check_calculation(bulk_calculation, deformation_calculation):
            deformation_calculations.append(deformation_calculation)

    sga = SpacegroupAnalyzer(bulk_structure, symprec=symprec)
    spg_symbol = unicodeify_spacegroup(sga.get_space_group_symbol())
    spg_number = sga.get_space_group_number()
    click.echo("\nSpacegroup: {} ({})".format(spg_symbol, spg_number))

    strain_mapping = get_strain_mapping(bulk_structure, deformation_calculations)
    click.echo("\nFound {} strains:".format(len(strain_mapping)))
    fmt_strain = get_formatted_tensors(strain_mapping.keys())
    click.echo("  - " + "\n  - ".join(fmt_strain))

    strain_mapping = get_symmetrized_strain_mapping(
        bulk_structure, strain_mapping, symprec=symprec
    )
    click.echo("\nAfter symmetrization found {} strains:".format(len(strain_mapping)))
    fmt_strain = get_formatted_tensors(strain_mapping.keys())
    click.echo("  - " + "\n  - ".join(fmt_strain))

    if not strain_coverage_ok(list(strain_mapping.keys())):
        click.echo("\nERROR: Strains do not cover full tensor, check calculations")
        sys.exit()

    click.echo("\nCalculating deformation potentials")
    bulk_calculation["bandstructure"] = expand_bandstructure(
        bulk_calculation["bandstructure"], symprec=symprec
    )
    deformation_potentials = calculate_deformation_potentials(
        bulk_calculation, strain_mapping
    )

    print_deformation_summary(bulk_calculation["bandstructure"], deformation_potentials)

    if "bands" in kwargs and kwargs["bands"] is not None:
        ibands = parse_ibands(kwargs["bands"])
    else:
        ibands = get_ibands(energy_cutoff, bulk_calculation["bandstructure"])

    echo_ibands(ibands, bulk_calculation["bandstructure"].is_spin_polarized)
    click.echo("")
    deformation_potentials = extract_bands(deformation_potentials, ibands)

    kpoints = get_kpoints_from_bandstructure(bulk_calculation["bandstructure"])
    filename = write_deformation_potentials(
        deformation_potentials, kpoints, bulk_structure, filename=kwargs["output"]
    )
    click.echo("\nDeformation potentials written to {}".format(filename))


def check_calculation(bulk_calculation, deformation_calculation):
    from amset.deformation.potentials import get_mesh_from_band_structure

    bulk_mesh, bulk_is_shifted = get_mesh_from_band_structure(
        bulk_calculation["bandstructure"]
    )
    bulk_species = tuple(bulk_calculation["bandstructure"].structure.species)
    bulk_is_metal = bulk_calculation["bandstructure"].is_metal()

    mesh, is_shifted = get_mesh_from_band_structure(
        deformation_calculation["bandstructure"]
    )
    if mesh != bulk_mesh or is_shifted != bulk_is_shifted:
        raise RuntimeError(
            "Calculations were not performed using the same k-point "
            "mesh\n{} != {}".format(mesh, bulk_mesh)
        )

    species = tuple(deformation_calculation["bandstructure"].structure.species)
    if species != bulk_species:
        raise RuntimeError("Calculations were performed using different structures")
    is_metal = deformation_calculation["bandstructure"].is_metal()

    if is_metal != bulk_is_metal:
        click.echo(
            "Bulk structure is {} whereas deformed structure is {}.\nSkipping "
            "deformation.".format(_metal_str[bulk_is_metal], _metal_str[is_metal])
        )
        return False
    else:
        return True


def print_deformation_summary(bandstructure, deformation_potentials):
    if bandstructure.is_metal():
        return
    else:
        cbm = bandstructure.get_cbm()
        vbm = bandstructure.get_vbm()
        click.echo("\nValence band maximum:")
        print_band_edge_information(bandstructure, vbm, deformation_potentials)

        click.echo("Conduction band minimum:")
        print_band_edge_information(bandstructure, cbm, deformation_potentials)


def print_band_edge_information(bandstructure, band_edge, deformation_potentials):
    for spin, spin_band_idxs in band_edge["band_index"].items():
        for b_idx in spin_band_idxs:
            for k_idx in band_edge["kpoint_index"]:
                kpoint = bandstructure.kpoints[k_idx].frac_coords
                edge_deform = deformation_potentials[spin][b_idx, k_idx]

                if len(deformation_potentials) == 2:
                    click.echo("  - spin {}:".format(spin.name))

                str_kpoint = _kpt_str.format(k=kpoint)
                click.echo("  - band: {:4d}  k-point: {}".format(b_idx, str_kpoint))
                click.echo("  - deformation potential:")
                click.echo(_tensor_str.format(*edge_deform.ravel()))
                click.echo("")



from typing import Tuple

import numpy as np
from pymatgen import Structure
from pymatgen.electronic_structure.bandstructure import BandStructure

from amset.constants import angstrom_to_bohr, bohr_to_angstrom, defaults

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

from pymatgen.io.vasp import Vasprun


def get_energy_cutoffs(
    energy_cutoff: float, band_structure: BandStructure
) -> Tuple[float, float]:
    if energy_cutoff and band_structure.is_metal():
        min_e = band_structure.efermi - energy_cutoff
        max_e = band_structure.efermi + energy_cutoff
    elif energy_cutoff:
        min_e = band_structure.get_vbm()["energy"] - energy_cutoff
        max_e = band_structure.get_cbm()["energy"] + energy_cutoff
    else:
        min_e = min(
            [band_structure.bands[spin].min() for spin in band_structure.bands.keys()]
        )
        max_e = max(
            [band_structure.bands[spin].max() for spin in band_structure.bands.keys()]
        )

    return min_e, max_e


def get_ibands(energy_cutoff: float, band_structure: BandStructure, return_idx=True):
    min_e, max_e = get_energy_cutoffs(energy_cutoff, band_structure)

    ibands = {}
    for spin, bands in band_structure.bands.items():
        ibands[spin] = np.any((bands > min_e) & (bands < max_e), axis=1)

        if return_idx:
            ibands[spin] = np.where(ibands[spin])[0]

    return ibands


def get_vb_idx(energy_cutoff: float, band_structure: BandStructure):
    if band_structure.is_metal():
        return None

    ibands = get_ibands(energy_cutoff, band_structure, return_idx=False)

    new_vb_idx = {}
    for spin, bands in band_structure.bands.items():
        spin_ibands = ibands[spin]

        # valence bands are all bands that contain energies less than efermi
        vbs = (bands < band_structure.efermi).any(axis=1)
        vb_idx = np.where(vbs)[0].max()

        # need to know the index of the valence band after discounting
        # bands during the interpolation. As ibands is just a list of
        # True/False, we can count the number of Trues included up to
        # and including the VBM to get the new number of valence bands
        new_vb_idx[spin] = sum(spin_ibands[: vb_idx + 1]) - 1

    return new_vb_idx


def get_velocities_from_outer_product(
    velocities_product, return_norm=False, symmetry_information=None
):
    # not recommended to give symmetry information AND return_norm=False
    # probably raise a warning
    velocities = {}

    for spin, spin_velocities_product in velocities_product.items():

        if symmetry_information:
            ir_kpoints_idx = symmetry_information["ir_kpoints_idx"]
            spin_velocities_product = spin_velocities_product[..., ir_kpoints_idx]

        v = np.sqrt(np.diagonal(spin_velocities_product, axis1=1, axis2=2))

        if return_norm:
            v = np.linalg.norm(v, axis=2)

        if symmetry_information:
            ir_to_full_idx = symmetry_information["ir_to_full_kpoint_mapping"]
            v = spin_velocities_product[..., ir_to_full_idx]

        velocities[spin] = v

    return velocities


def get_atomic_structure(structure):
    return Structure(
        structure.lattice.matrix * angstrom_to_bohr,
        structure.species,
        structure.frac_coords,
        site_properties=structure.site_properties,
    )


def get_angstrom_structure(structure):
    return Structure(
        structure.lattice.matrix * bohr_to_angstrom,
        structure.species,
        structure.frac_coords,
        site_properties=structure.site_properties,
    )


def get_band_structure(
    vasprun: Vasprun, zero_weighted: str = defaults["zero_weighted_kpoints"]
) -> BandStructure:
    """
    Get a band structure from a Vasprun object.

    This can ensure that if the calculation contains zero-weighted k-points then the
    weighted k-points will be discarded (helps with hybrid calculations).

    Also ensures that the Fermi level is set correctly.

    Args:
        vasprun: A vasprun object.
        zero_weighted: How to handle zero-weighted k-points if they are present in the
            calculation. Options are:
            - "keep": Keep zero-weighted k-points in the band structure.
            - "drop": Drop zero-weighted k-points, keeping only the weighted k-points.
            - "prefer": Drop weighted-kpoints if zero-weighted k-points are present
              in the calculation (useful for cheap hybrid calculations).

    Returns:
        A band structure.
    """
    # first check if Fermi level crosses a band
    k_idx = get_zero_weighted_kpoint_indices(vasprun, mode=zero_weighted)
    kpoints = np.array(vasprun.actual_kpoints)[k_idx]

    projections = {}
    eigenvalues = {}
    for spin, spin_eigenvalues in vasprun.eigenvalues.items():
        # discard weight and set shape nbands, nkpoints
        eigenvalues[spin] = spin_eigenvalues[k_idx, :, 0].transpose(1, 0)

        if vasprun.projected_eigenvalues:
            # is nkpoints, nbands, nion, norb; we need nbands, nkpoints, norb, nion
            spin_projections = vasprun.projected_eigenvalues[spin]
            projections[spin] = spin_projections[k_idx].transpose(1, 0, 3, 2)

    # finding the Fermi level is quite painful, as VASP can sometimes put it slightly
    # inside a band
    fermi_crosses_band = False
    for spin_eigenvalues in eigenvalues.values():
        eigs_below = np.any(spin_eigenvalues < vasprun.efermi, axis=1)
        eigs_above = np.any(spin_eigenvalues > vasprun.efermi, axis=1)
        if np.any(eigs_above & eigs_below):
            fermi_crosses_band = True

    # if the Fermi level crosses a band, the eigenvalue band properties is a more
    # reliable way to check whether this is a real effect
    bandgap, cbm, vbm, _ = vasprun.eigenvalue_band_properties

    if not fermi_crosses_band:
        # safe to use VASP fermi level
        efermi = vasprun.efermi
    elif fermi_crosses_band and bandgap == 0:
        # it is actually a metal
        efermi = vasprun.efermi
    else:
        # Set Fermi level half way between valence and conduction bands
        efermi = (cbm + vbm) / 2

    return BandStructure(
        kpoints,
        eigenvalues,
        vasprun.final_structure.lattice.reciprocal_lattice,
        efermi,
        structure=vasprun.final_structure,
        projections=projections,
    )


def get_zero_weighted_kpoint_indices(vasprun: Vasprun, mode: str) -> np.ndarray:
    """
    Get zero weighted k-point k-point indices from a vasprun.

    If the calculation doesn't contain zero-weighted k-points, then the indices of
    all the k-points will be returned. Alternatively, if the calculation contains
    a mix of weighted and zero-weighted k-points, then only the indices of the
    zero-weighted k-points will be returned.

    Args:
        vasprun:  A vasprun object.
        mode: How to handle zero-weighted k-points if they are present in the
            calculation. Options are:
            - "keep": Keep zero-weighted k-points in the band structure.
            - "drop": Drop zero-weighted k-points, keeping only the weighted k-points.
            - "prefer": Drop weighted-kpoints if zero-weighted k-points are present
              in the calculation (useful for cheap hybrid calculations).

    Returns:
        The indices of the valid k-points.
    """
    weights = np.array(vasprun.actual_kpoints_weights)
    is_zero_weight = weights == 0

    if mode not in ("prefer", "drop", "keep"):
        raise ValueError(f"Unrecognised zero-weighted k-point mode: {mode}")

    if mode == "prefer" and np.any(is_zero_weight):
        return np.where(is_zero_weight)[0]
    elif mode == "drop" and np.any(~is_zero_weight):
        return np.where(~is_zero_weight)[0]
    else:
        return np.arange(len(weights))



import numpy as np


def fd(e, mu, kb_t):
    """Compute Fermi-Dirac occupancies.

    Args:
        e: array of energies
        mu: single value of the chemical potential
        kb_t: thermal energy at the temperature of interest

    Returns:
        An array with the same shape as the "e" argument containing the average
        Fermi-Dirac occupancies for each energy level.
    """
    if kb_t == 0.0:
        delta = e - mu
        nruter = np.where(delta < 0.0, 1.0, 0.0)
        nruter[np.isclose(delta, 0.0)] = 0.5
    else:
        x = np.asarray((e - mu) / kb_t)
        nruter = 1.0 / (np.exp(x) + 1.0)
    return nruter


def dfddx(x):
    """Compute the derivative of the Fermi-Dirac occupancies with respect to
    the recentered and scaled energy (e - mu) / kBT

    Args:
        x: recentered and scaled energies at which to compute the derivative

    Returns:
        An array with the same shape as x containing the derivatives.
    """
    c = np.cosh(0.5 * np.asarray(x))
    nruter = -0.25 / c / c
    return nruter


def dfdde(e, mu, kb_t):
    """Compute the derivative of Fermi-Dirac occupancies with respect to
    the energy.

    Args:
        e: array of energies
        mu: single value of the chemical potential
        kb_t: thermal energy at the temperature of interest

    Returns:
        An array with the same shape as the "e" argument containing the
        derivatives of the average Fermi-Dirac occupancies for each energy
        level.
    """
    factor = (e - mu) / kb_t
    dfde = dfddx(factor) / kb_t
    return dfde
